###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Unit tests for Workflow Modules."""
# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import unittest
import os
import copy
import shutil

from unittest.mock import MagicMock, patch

from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from DIRAC.Resources.Catalog.test.mock_FC import fc_mock

from DIRAC import gLogger
from DIRAC.RequestManagementSystem.Client.Request import Request

# mocks
from LHCbDIRAC.Workflow.Modules.mock_Commons import (
    prod_id,
    prod_job_id,
    wms_job_id,
    workflowStatus,
    stepStatus,
    step_id,
    step_number,
    step_commons,
    wf_commons,
    rc_mock,
)
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock

# sut
from LHCbDIRAC.Workflow.Modules.FailoverRequest import FailoverRequest
from LHCbDIRAC.Workflow.Modules.RemoveInputData import RemoveInputData
from LHCbDIRAC.Workflow.Modules.UserJobFinalization import UserJobFinalization
from LHCbDIRAC.Workflow.Modules.StepAccounting import StepAccounting
from LHCbDIRAC.Workflow.Modules.UploadLogFile import UploadLogFile
from LHCbDIRAC.Workflow.Modules.FileUsage import FileUsage
from LHCbDIRAC.Workflow.Modules.CreateDataFile import CreateDataFile

getDestinationSEListMock = MagicMock()
getDestinationSEListMock.return_value = []
getDestinationSEListMockCNAF = MagicMock()
getDestinationSEListMockCNAF.return_value = ["CNAF"]


class ModulesTestCase(unittest.TestCase):
    """Base class for the Modules test cases."""

    def setUp(self):
        gLogger.setLevel("DEBUG")
        self.maxDiff = None

        self.jsu_mock = MagicMock()
        self.jsu_mock.setJobApplicationStatus.return_value = {"OK": True, "Value": ""}

        self.jsu_mock = MagicMock()
        self.jsu_mock.setJobApplicationStatus.return_value = {"OK": True, "Value": ""}

        self.ft_mock = MagicMock()
        self.ft_mock.transferAndRegisterFile.return_value = {"OK": True, "Value": {"uploadedSE": ""}}
        self.ft_mock.transferAndRegisterFileFailover.return_value = {"OK": True, "Value": {}}
        self.ft_mock.request = rc_mock
        self.ft_mock.FileCatalog = fc_mock

        self.nc_mock = MagicMock()
        self.nc_mock.sendMail.return_value = {"OK": True, "Value": ""}

        self.xf_o_mock = MagicMock()
        self.xf_o_mock.inputFileStats = {"a": 1, "b": 2}
        self.xf_o_mock.outputFileStats = {"a": 1, "b": 2}
        self.xf_o_mock.analyse.return_value = True

        self.jobStep_mock = MagicMock()
        self.jobStep_mock.commit.return_value = {"OK": True, "Value": ""}
        self.jobStep_mock.setValuesFromDict.return_value = {"OK": True, "Value": ""}
        self.jobStep_mock.checkValues.return_value = {"OK": True, "Value": ""}

    def tearDown(self):
        for fileProd in [
            "appLog",
            "foo.txt",
            "aaa.Bhadron.dst",
            "bbb.Calibration.dst",
            "bar.py",
            "aLongLog.log",
            "bookkeeping_123_00000456_321.xml",
            "aLongLog.log.gz",
            "ccc.charm.mdst",
            "ccc.charm.mdst",
            "prova.txt",
            "aLog.log",
            "BAR.txt",
            "FooBAR.ext.txt",
            "foo_1.txt",
            "bar_2.py",
            "bar.txt",
            "ErrorLogging_Step1_coredump.log",
            "123_00000456_request.xml",
            "lfn1",
            "lfn2",
            "XMLSummaryFile",
            "aaa.bhadron.dst",
            "bbb.calibration.dst",
            "ProductionOutputData",
            "data.py",
            "123_00000456_request.json",
            "00000123_00000456.tar",
            "someOtherDir",
            "DISABLE_WATCHDOG_CPU_WALLCLOCK_CHECK",
            "myfoo.blah",
            "prodConf_someApp__.py",
            "prodConf_someApp___.py",
        ]:
            try:
                os.remove(fileProd)
            except OSError:
                continue

        for directory in ["./job", "job"]:
            try:
                shutil.rmtree(directory)
            except Exception:
                continue


# FailoverRequest.py
@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
class FailoverRequestSuccess(ModulesTestCase):
    #################################################

    def test_execute(self, _patch):
        fr = FailoverRequest(bkClient=bkc_mock, dm=dm_mock)
        fr.jobType = "merge"
        fr.stepInputData = ["foo", "bar"]
        fr.requestValidator = MagicMock()

        # no errors, no input data
        for wf_cs in copy.deepcopy(wf_commons):
            for s_cs in step_commons:
                self.assertTrue(
                    fr.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )


# RemoveInputData.py
@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
class RemoveInputDataSuccess(ModulesTestCase):
    #################################################

    def test_execute(self, _patch):
        rid = RemoveInputData(bkClient=bkc_mock, dm=dm_mock)
        # no errors, no input data
        for wf_cs in copy.deepcopy(wf_commons):
            if "InputData" in wf_cs:
                continue
            for s_cs in step_commons:
                self.assertTrue(
                    rid.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )

        # no errors, input data
        for wf_cs in copy.deepcopy(wf_commons):
            if "InputData" not in wf_cs:
                continue
            for s_cs in step_commons:
                self.assertTrue(
                    rid.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )


# StepAccounting.py
@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
class StepAccountingSuccess(ModulesTestCase):
    #################################################

    def test_execute(self, _patch):
        sa = StepAccounting(bkClient=bkc_mock, dm=dm_mock)
        sa.jobType = "merge"
        sa.stepInputData = ["foo", "bar"]
        sa.siteName = "DIRAC.Test.ch"

        for wf_cs in copy.deepcopy(wf_commons):
            for s_cs in step_commons:
                self.assertTrue(
                    sa.execute(
                        prod_id,
                        prod_job_id,
                        wms_job_id,
                        workflowStatus,
                        stepStatus,
                        wf_cs,
                        s_cs,
                        step_number,
                        step_id,
                        self.jobStep_mock,
                        self.xf_o_mock,
                    )["OK"]
                )


# UploadLogFile.py
@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
class UploadLogFileSuccess(ModulesTestCase):
    @patch("LHCbDIRAC.Workflow.Modules.UploadLogFile.getDestinationSEList", side_effect=getDestinationSEListMockCNAF)
    def test_execute(self, _patch, _patch1):
        # no errors, no input data
        #    for wf_commons in copy.deepcopy( wf_commons ):
        #      for step_commons in step_commons:
        #        self.assertTrue( ulf.execute( prod_id, prod_job_id, wms_job_id,
        #                                           workflowStatus, stepStatus,
        #                                           wf_commons, step_commons,
        #                                           step_number, step_id,
        #                                           dm_mock, self.ft_mock,
        #                                           bkc_mock )['OK'] )

        # putStorageDirectory returns False

        rm_mock = copy.deepcopy(dm_mock)
        rm_mock.putStorageDirectory.return_value = {"OK": False, "Message": "bih"}
        ft_mock = copy.deepcopy(self.ft_mock)
        ulf = UploadLogFile(bkClient=bkc_mock, dm=dm_mock)
        ulf.request = Request()
        ulf.failoverTransfer = ft_mock
        for wf_cs in copy.deepcopy(wf_commons):
            for s_cs in step_commons:
                self.assertTrue(
                    ulf.execute(prod_id, prod_job_id, 0, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id)[
                        "OK"
                    ]
                )
            # self.assertTrue( ulf.finalize( rm_mock, self.ft_mock )['OK'] )

    @patch("LHCbDIRAC.Workflow.Modules.UploadLogFile.getDestinationSEList", side_effect=getDestinationSEListMockCNAF)
    def test__uploadLogToFailoverSE(self, _patch, _patch1):
        open("foo.txt", "w").close()
        tarFileName = "foo.txt"
        ulf = UploadLogFile(bkClient=bkc_mock, dm=dm_mock)
        ulf.request = Request()
        ulf.failoverTransfer = self.ft_mock
        ulf.logLFNPath = "/an/lfn/foo.txt"
        ulf.failoverSEs = ["SE1", "SE2"]
        ulf._uploadLogToFailoverSE(tarFileName)

    @patch("LHCbDIRAC.Workflow.Modules.UploadLogFile.getDestinationSEList", side_effect=getDestinationSEListMockCNAF)
    def test__determinRelevantFiles(self, _patch, _patch1):
        for fileN in ["foo.txt", "aLog.log", "aLongLog.log", "aLongLog.log.gz"]:
            try:
                os.remove(fileN)
            except OSError:
                continue

        open("foo.txt", "w").close()
        open("bar.py", "w").close()
        open("aLog.log", "w").close()
        ulf = UploadLogFile(bkClient=bkc_mock, dm=dm_mock)
        res = ulf._determineRelevantFiles()
        self.assertTrue(res["OK"])
        expected = ["foo.txt", "aLog.log"]
        if "pylint.txt" in os.listdir("."):
            expected.append("pylint.txt")
        if "nosetests.xml" in os.listdir("."):
            expected.append("nosetests.xml")
        self.assertTrue(set(res["Value"]) >= set(expected))

        fd = open("aLongLog.log", "w")
        for _x in range(2500):
            fd.writelines(
                "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do "
                "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim "
                "ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut "
                "aliquip ex ea commodo consequat. Duis aute irure dolor in "
                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla "
                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in "
                "culpa qui officia deserunt mollit anim id est laborum"
            )
        fd.close()
        res = ulf._determineRelevantFiles()
        self.assertTrue(res["OK"])
        expected = ["foo.txt", "aLog.log", "aLongLog.log"]
        if "pylint.txt" in os.listdir("."):
            expected.append("pylint.txt")
        if "nosetests.xml" in os.listdir("."):
            expected.append("nosetests.xml")
        self.assertTrue(set(res["Value"]) >= set(expected))

        open("foo.txt", "w").close()
        fd = open("aLongLog.log", "w")
        for _x in range(2500):
            fd.writelines(
                "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do "
                "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim "
                "ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut "
                "aliquip ex ea commodo consequat. Duis aute irure dolor in "
                "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla "
                "pariatur. Excepteur sint occaecat cupidatat non proident, sunt in "
                "culpa qui officia deserunt mollit anim id est laborum"
            )
        fd.close()
        open("bar.py", "w").close()
        res = ulf._determineRelevantFiles()
        expected = ["foo.txt", "aLog.log", "aLongLog.log"]
        if "pylint.txt" in os.listdir("."):
            expected.append("pylint.txt")
        if "nosetests.xml" in os.listdir("."):
            expected.append("nosetests.xml")
        self.assertTrue(res["OK"])
        self.assertTrue(set(res["Value"]) >= set(expected))


# UserJobFinalization.py
@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
class UserJobFinalizationSuccess(ModulesTestCase):
    def test_execute(self, _patch):
        ujf = UserJobFinalization(bkClient=bkc_mock, dm=dm_mock)
        ujf.bkClient = bkc_mock
        ujf.failoverTransfer = self.ft_mock
        ujf.requestValidator = MagicMock()
        ujf.requestValidator.validate.return_value = {"OK": True}

        # no errors, no input data
        for wf_cs in copy.deepcopy(wf_commons):
            wf_cs["TotalSteps"] = step_number
            for s_cs in step_commons:
                self.assertTrue(
                    ujf.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )

        for wf_cs in copy.deepcopy(wf_commons):
            wf_cs["TotalSteps"] = step_number
            for s_cs in step_commons:
                wf_cs["UserOutputData"] = ["i1", "i2"]
                wf_cs["UserOutputSE"] = ["MySE"]
                wf_cs["OwnerName"] = "fstagni"
                open("i1", "w").close()
                open("i2", "w").close()
                self.assertTrue(
                    ujf.execute(
                        prod_id,
                        prod_job_id,
                        wms_job_id,
                        workflowStatus,
                        stepStatus,
                        wf_cs,
                        s_cs,
                        step_number,
                        step_id,
                        orderedSEs=["MySE1", "MySE2"],
                    )["OK"]
                )
            os.remove("i1")
            os.remove("i2")

    @patch("LHCbDIRAC.Workflow.Modules.UserJobFinalization.getDestinationSEList", side_effect=getDestinationSEListMock)
    def test__getOrderedSEsList(self, _patch, _patched):
        ujf = UserJobFinalization(bkClient=bkc_mock, dm=dm_mock)

        ujf.userOutputSE = ["userSE"]
        res = ujf._getOrderedSEsList()
        self.assertEqual(res, ["userSE"])

        ujf.defaultOutputSE = ["CERN"]
        res = ujf._getOrderedSEsList()
        self.assertEqual(res, ["userSE", "CERN"])

    @patch(
        "LHCbDIRAC.Workflow.Modules.UserJobFinalization.getDestinationSEList", side_effect=getDestinationSEListMockCNAF
    )
    def test__getOrderedSEsListCNAF(self, _patch, _patched):
        ujf = UserJobFinalization(bkClient=bkc_mock, dm=dm_mock)
        res = ujf._getOrderedSEsList()
        self.assertEqual(res, ["CNAF"])


# FileUsage.py
@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
class FileUsageSuccess(ModulesTestCase):
    #################################################

    def test_execute(self, _patch):
        # no errors, no input files to report
        wfStatus = copy.deepcopy(workflowStatus)
        wfC = copy.deepcopy(wf_commons)
        for wf_cs in wfC:  # copy.deepcopy( wf_commons ):
            for s_cs in step_commons:
                fu = FileUsage(bkClient=bkc_mock, dm=dm_mock)
                self.assertTrue(
                    fu.execute(
                        prod_id, prod_job_id, wms_job_id, wfStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )
        wfC = copy.deepcopy(wf_commons)
        for wf_cs in wfC:  # copy.deepcopy( wf_commons ):
            wf_cs["ParametricInputData"] = [
                "LFN:/lhcb/data/2010/EW.DST/00008380/0000/00008380_00000287_1.ew.dst",
                "LFN:/lhcb/data/2010/EW.DST/00008380/0000/00008380_00000285_1.ew.dst",
                "LFN:/lhcb/data/2010/EW.DST/00008380/0000/00008380_00000281_1.ew.dst",
            ]
            for s_cs in step_commons:
                fu = FileUsage(bkClient=bkc_mock, dm=dm_mock)
                self.assertTrue(
                    fu.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )
        wfC = copy.deepcopy(wf_commons)
        for wf_cs in wfC:  # copy.deepcopy( wf_commons ):
            wf_cs["ParametricInputData"] = [
                "LFN:/lhcb/data/2010/EW.DST/00008380/0000/00008380_00000287_1.ew.dst",
                "LFN:/lhcb/data/2010/EW.DST/00008380/0000/00008380_00000285_1.ew.dst",
                "LFN:/lhcb/data/2010/PIPPO/00008380/0000/00008380_00000281_1.pippo.dst",
            ]
            for s_cs in step_commons:
                fu = FileUsage(bkClient=bkc_mock, dm=dm_mock)
                self.assertTrue(
                    fu.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )

        # workflow status not ok
        wfC = copy.deepcopy(wf_commons)
        for wf_cs in wfC:  # copy.deepcopy( wf_commons ):
            wfStatus = {"OK": False, "Message": "Mess"}
            for s_cs in step_commons:
                fu = FileUsage(bkClient=bkc_mock, dm=dm_mock)
                self.assertTrue(
                    fu.execute(
                        prod_id, prod_job_id, wms_job_id, wfStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )


# FileUsage.py
@patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator", side_effect=MagicMock())
class CreateDataFileSuccess(ModulesTestCase):
    #################################################

    def test_execute(self, _patch):
        cdf = CreateDataFile(bkClient=bkc_mock, dm=dm_mock)
        cdf.jobType = "merge"
        cdf.stepInputData = ["foo", "bar"]

        for wf_cs in copy.deepcopy(wf_commons):
            for s_cs in step_commons:
                self.assertTrue(
                    cdf.execute(
                        prod_id, prod_job_id, wms_job_id, workflowStatus, stepStatus, wf_cs, s_cs, step_number, step_id
                    )["OK"]
                )


if __name__ == "__main__":
    unittest.main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import os
import time
from typing import Any
from collections.abc import Callable

from DIRAC import gLogger, gConfig
from DIRAC.ConfigurationSystem.Client.PathFinder import getDatabaseSection
from DIRAC.Core.Utilities.ReturnValues import DReturnType
from DIRAC.FrameworkSystem.private.standardLogging.LoggingRoot import LoggingRoot

from .OracleDB import OracleDB
from .LegacyOracleBookkeepingDB import LegacyOracleBookkeepingDB
from .NewOracleBookkeepingDB import NewOracleBookkeepingDB


DCallable = Callable[..., DReturnType[Any]]


class OracleBookkeepingDB:
    """Proxy class which routes calls to the legacy or new OracleBookkeepingDB implementation.

    This class has three modes of operation:

    * By default it call methods in ``LegacyOracleBookkeepingDB`` by default and only use ``NewOracleBookkeepingDB``
        if the requested method is not implemented in ``LegacyOracleBookkeepingDB``.
    * If ``LHCBDIRAC_BOOKKEEPING_DO_COMPARE`` is set, it will call both implementations and compare the results for
        methods which are implemented in both and deemed safe to compare. If the results differ, it will log a warning.
        The safety of calling a method is determined by the ``DUPLICATION_SAFE_METHODS`` variable.
    * The ``LHCBDIRAC_BOOKKEEPING_METHODS_PREFER_NEW`` environment variable is a comma separated list that can be used
        to force the proxy to prefer the new implementation for specific methods.
    """

    def __init__(self, *, host: str = None, username: str = None, password: str = None, **kwargs):
        self.log = gLogger.getSubLogger("ProxyOracleBookkeepingDB")
        self._legacydb = LegacyOracleBookkeepingDB(
            dbR=OracleDB(**kwargs, **_init_kwargs(with_write=False, username=username, password=password, host=host)),
            dbW=OracleDB(**kwargs, **_init_kwargs(with_write=True, username=username, password=password, host=host)),
        )
        self._newdb = NewOracleBookkeepingDB(
            dbR=OracleDB(**kwargs, **_init_kwargs(with_write=False, username=username, password=password, host=host)),
            dbW=OracleDB(**kwargs, **_init_kwargs(with_write=True, username=username, password=password, host=host)),
        )

    def __dir__(self) -> list[str]:
        return list(set(dir(self._legacydb)) | set(dir(self._newdb)))

    def __getattr__(self, name: str) -> DCallable:
        legacy_impl = getattr(self._legacydb, name, None)
        new_impl = getattr(self._newdb, name, None)
        if legacy_impl is None and new_impl is None:
            raise AttributeError(f"{self.__name!r} object has no attribute {name!r}")

        # If we only find it on one of the implementations, just return that callable
        if legacy_impl is None:
            return new_impl
        if new_impl is None:
            return legacy_impl

        # Check which implementation we prefer according to the environment variable
        if name in METHODS_PREFER_NEW:
            self.log.debug("Preferring new implementation", f"for {name}")
            preferred_impl = new_impl
            reference_impl = legacy_impl
        else:
            self.log.debug("Preferring old implementation", f"for {name}")
            preferred_impl = legacy_impl
            reference_impl = new_impl

        if not DO_COMPARE:
            return preferred_impl

        # Check if the method is safe for duplication (i.e. read-only)
        if name not in DUPLICATION_SAFE_METHODS:
            self.log.debug("Method is not safe for duplication", f"for {name}")
            return preferred_impl

        return ProxyMethod(self.log, preferred_impl, reference_impl)


def _init_kwargs(*, with_write: bool, username: str = None, password: str = None, host: str = None) -> dict[str, str]:
    """Get the connection keyword arguments for the OracleDB constructor.

    For any parameters which are not specified, the values are read from
    ``gConfig``. If ``username`` is not given, the ``with_write`` parameter
    determines whether the ``LHCbDIRACBookkeepingUser`` (read-only) or
    ``LHCbDIRACBookkeepingServer`` (read-write) configuration option is used.
    """
    cs_path = getDatabaseSection("Bookkeeping", "BookkeepingDB")

    if host is None:
        result = gConfig.getOption(cs_path + "/LHCbDIRACBookkeepingTNS")
        if not result["OK"]:
            raise ValueError("Failed to get the configuration parameters: LHCbDIRACBookkeepingTNS")
        host = result["Value"]

    if username is None:
        if with_write:
            result = gConfig.getOption(cs_path + "/LHCbDIRACBookkeepingServer")
        else:
            result = gConfig.getOption(cs_path + "/LHCbDIRACBookkeepingUser")
        if not result["OK"]:
            raise ValueError("Failed to get the configuration parameter for username", f"for {with_write=}")
        username = result["Value"]

    if password is None:
        result = gConfig.getOption(cs_path + "/LHCbDIRACBookkeepingPassword")
        if not result["OK"]:
            raise ValueError("Failed to get the configuration parameters: LHCbDIRACBookkeepingPassword")
        password = result["Value"]

    return dict(
        userName=username,
        password=password,
        tnsEntry=host,
        confDir=gConfig.getValue(cs_path + "/LHCbDIRACBookkeepingConfDir", ""),
        mode=gConfig.getValue(cs_path + "/LHCbDIRACBookkeepingMode", ""),
    )


class ProxyMethod:
    """Callable class which calls the two bookkeeping database implementations and logs the results."""

    def __init__(self, log: LoggingRoot, preferred_impl: DCallable | None, reference_impl: DCallable | None):
        self.log = log
        self.preferred_impl = preferred_impl
        self.reference_impl = reference_impl

    def __call__(self, *args, **kwargs):
        start = time.perf_counter()
        result = self.preferred_impl(*args, **kwargs)
        preferred_duration = time.perf_counter() - start

        start = time.perf_counter()
        reference = self.reference_impl(*args, **kwargs)
        reference_duration = time.perf_counter() - start

        error_message = None
        if {result["OK"], reference["OK"]} == {True, False}:
            error_message = "One implementation errored while the other did not"
        elif result["OK"]:
            if self.name in USE_SET_COMPARISON:
                try:
                    result_set = set(result["Value"])
                    reference_set = set(reference["Value"])
                except Exception as e:
                    error_message = f"Error while comparing sets: {e}"
                else:
                    if result_set != reference_set:
                        error_message = "Legacy and new methods returned different results"
                    if len(result_set) != len(result["Value"]):
                        error_message = "Legacy method returned duplicate results"
                    if len(reference_set) != len(reference["Value"]):
                        error_message = "New method returned duplicate results"
            elif result != reference:
                error_message = "Legacy and new methods returned different results"

        if error_message is None:
            diff_duration = preferred_duration - reference_duration
            diff_desc = "faster" if diff_duration > 0 else "slower"
            diff_rel = abs(diff_duration) / preferred_duration
            varmsg = (
                f"for {self.name}, reference was {abs(diff_duration):.4f} seconds {diff_desc} ({diff_rel:.1%}) "
                f"with {args=} {kwargs=}"
            )
            if diff_rel > 0.5 and diff_duration < -1:
                self.log.warn("Legacy and new methods matched but reference is notably slower", varmsg)
            else:
                self.log.info("Legacy and new methods matched", varmsg)
        else:
            msgresult = str(result)[:1000] if result["OK"] else result
            msgreference = str(reference)[:1000] if reference["OK"] else reference
            varmsg = (
                f"for {self.name} with {args=} {kwargs=}:\n"
                f"{self.preferred_name} gave {msgresult!r} in {preferred_duration:.4f} seconds\n"
                f"{self.reference_name} gave {msgreference!r} in {reference_duration:.4f} seconds"
            )
            self.log.warn(error_message, varmsg)
            if FAIL_ON_DIFFERENCE:
                raise BookkeepingResultMismatch(f"{error_message} {varmsg}")

        return result

    @property
    def name(self) -> str:
        return self.preferred_impl.__name__

    @property
    def preferred_name(self) -> str:
        return self.preferred_impl.__self__.__class__.__name__

    @property
    def reference_name(self) -> str:
        return self.reference_impl.__self__.__class__.__name__


class BookkeepingResultMismatch(Exception):
    pass


METHODS_PREFER_NEW = os.environ.get("LHCBDIRAC_BOOKKEEPING_METHODS_PREFER_NEW", "").split(",")
FAIL_ON_DIFFERENCE = os.environ.get("LHCBDIRAC_BOOKKEEPING_FAIL_ON_DIFFERENCE", "") == "True"
DO_COMPARE = os.environ.get("LHCBDIRAC_BOOKKEEPING_DO_COMPARE") == "True"
# Some methods don't have a well defined order of results, so we need to compare them as sets
USE_SET_COMPARISON = {
    "getFiles",
    "getFilesWithMetadata",
    "getVisibleFilesWithMetadata",
}
DUPLICATION_SAFE_METHODS = [
    # "addProcessing",
    # "addProduction",
    # "addProductionSteps",
    # "addReplica",
    # "bulkgetIDsFromFilesTable",
    # "bulkinsertEventType",
    # "bulkJobInfo",
    # "bulkupdateEventType",
    # "bulkupdateFileMetaData",
    # "checkEventType",
    # "checkfile",
    # "checkFileTypeAndVersion",
    # "checkProcessingPassAndSimCond",
    # "deleteCertificationData",
    # "deleteDataTakingCondition",
    # "deleteFile",
    # "deleteFiles",
    # "deleteInputFiles",
    # "deleteJob",
    # "deleteProductionsContainer",
    # "deleteSimulationConditions",
    # "deleteStep",
    # "deleteStepContainer",
    # "exists",
    # "existsTag",
    # "fixRunLuminosity",
    # "getAvailableConfigNames",
    # "getAvailableConfigurations",
    # "getAvailableDataQuality",
    # "getAvailableEventTypes",
    "getAvailableFileTypes",
    # "getAvailableProductions",
    # "getAvailableRuns",
    # "getAvailableSteps",
    # "getAvailableTags",
    # "getAvailableTagsFromSteps",
    # "getConditions",
    # "getConfigsAndEvtType",
    # "getConfigVersions",
    # "getDataTakingCondDesc",
    # "getDataTakingCondId",
    # "getDirectoryMetadata",
    # "getEventTypes",
    # "getFileAncestorHelper",
    # "getFileAncestors",
    # "getFileCreationLog",
    "getFileDescendents",
    # "getFileHistory",
    # "getFileMetadata",
    "getFiles",
    # "getFilesForGUID",
    "getFilesSummary",
    "getFilesWithMetadata",
    # "getFileTypes",
    # "getFileTypeVersion",
    # "getInputFiles",
    # "getJobInfo",
    # "getJobInformation",
    # "getJobInputOutputFiles",
    # "getLimitedFiles",
    # "getListOfFills",
    # "getListOfRuns",
    # "getMoreProductionInformations",
    # "getNbOfJobsBySites",
    # "getNbOfRawFiles",
    # "getOutputFiles",
    # "getProcessingPass",
    # "getProcessingPassId",
    # "getProcessingPassSteps",
    # "getProductionFiles",
    # "getProductionFilesBulk",
    # "getProductionFilesForWeb",
    # "getProductionFilesStatus",
    # "getProductionInformation",
    # "getProductionNbOfEvents",
    # "getProductionNbOfFiles",
    # "getProductionNbOfJobs",
    # "getProductionOutputFileTypes",
    # "getProductionProcessedEvents",
    # "getProductionProcessingPass",
    # "getProductionProcessingPassID",
    # "getProductionProcessingPassSteps",
    # "getProductionProducedEvents",
    # "getProductions",
    # "getProductionsFromView",
    # "getProductionSimulationCond",
    # "getProductionSizeOfFiles",
    # "getProductionSummary",
    # "getProductionSummaryFromView",
    # "getRunAndProcessingPass",
    # "getRunAndProcessingPassDataQuality",
    # "getRunConfigurationsAndDataTakingCondition",
    # "getRunFiles",
    # "getRunFilesDataQuality",
    # "getRunInformation",
    # "getRunInformations",
    # "getRunNbAndTck",
    # "getRunNumber",
    # "getRunProcessingPass",
    # "getRuns",
    # "getRunsForAGivenPeriod",
    # "getRunsForFill",
    # "getRunsGroupedByDataTaking",
    # "getRunStatus",
    # "getRuntimeProjects",
    # "getRunWithProcessingPassAndDataQuality",
    # "getSimConditions",
    # "getSimulationConditions",
    # "getStepIdandNameForRUN",
    # "getStepInputFiles",
    # "getStepOutputFiles",
    # "getSteps",
    # "getStepsMetadata",
    # "getTCKs",
    "getVisibleFilesWithMetadata",
    # "insertDataTakingCond",
    # "insertDataTakingCondDesc",
    # "insertEventTypes",
    # "insertFileTypes",
    # "insertInputFile",
    # "insertJob",
    # "insertOutputFile",
    # "insertProductionOutputFiletypes",
    # "insertproductionscontainer",
    # "insertRunStatus",
    # "insertRuntimeProject",
    # "insertSimConditions",
    # "insertStep",
    # "insertTag",
    # "removeReplica",
    # "removeRuntimeProject",
    # "renameFile",
    # "setFileDataQuality",
    # "setFilesInvisible",
    # "setFilesVisible",
    # "setProductionDataQuality",
    # "setRunAndProcessingPassDataQuality",
    # "setRunDataQuality",
    # "setRunStatusFinished",
    # "setStepInputFiles",
    # "setStepOutputFiles",
    # "updateEventType",
    # "updateFileMetaData",
    # "updateProductionOutputfiles",
    # "updateReplicaRow",
    # "updateRuntimeProject",
    # "updateSimulationConditions",
    # "updateStep",
]



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""LHCbDIRAC.ResourceStatusSystem.Policy.JobEfficiencyPolicy.

JobWebSummaryEfficiencyPolicy.__bases__:
  DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase.PolicyBase
"""
from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase


class JobWebSummaryEfficiencyPolicy(PolicyBase):
    """The JobEfficiencyPolicy class is a policy that checks the efficiency of
    the jobs according to what is on WMS.

    Evaluates the JobEfficiency results given by the JobCommand.JobCommand
    """

    @staticmethod
    def _evaluate(commandResult):
        """Evaluate policy on jobs stats, using args (tuple).

        :returns:
          {
            'Status':Unknown|Active|Probing|Bad,
            'Reason':'JobsEff:Good|JobsEff:Fair|JobsEff:Poor|JobsEff:Bad|JobsEff:Idle',
          }
        """

        result = {"Status": None, "Reason": None}

        if not commandResult["OK"]:
            result["Status"] = "Error"
            result["Reason"] = commandResult["Message"]
            return S_OK(result)

        commandResult = commandResult["Value"]

        if not commandResult:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decision"
            return S_OK(result)

        # The command returns a list of dictionaries, with only one if thre is something,
        # otherwise an empty list.
        commandResult = commandResult[0]

        if "Status" not in commandResult:
            result["Status"] = "Error"
            result["Reason"] = '"Status" key missing'
            return S_OK(result)

        if "Efficiency" not in commandResult:
            result["Status"] = "Error"
            result["Reason"] = '"Efficiency" key missing'
            return S_OK(result)

        status = commandResult["Status"]
        efficiency = commandResult["Efficiency"]

        if status == "Good":
            result["Status"] = "Active"
        elif status == "Fair":
            result["Status"] = "Active"
        elif status == "Poor":
            result["Status"] = "Degraded"
        elif status == "Idle":
            result["Status"] = "Unknown"
        elif status == "Bad":
            result["Status"] = "Banned"
        else:
            result["Status"] = "Error"
            result["Reason"] = f'Unknown status "{status}"'
            return S_OK(result)

        result["Reason"] = f"Jobs Efficiency: {efficiency} with status {status}"

        return S_OK(result)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Moving toward a templates-less system."""
import ast

from DIRAC.Core.Base.Script import Script

Script.parseCommandLine()

from DIRAC import gConfig, gLogger, exit as DIRACexit
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest

gLogger = gLogger.getSubLogger("LaunchingRequest_run.py")
currentSetup = gConfig.getValue("DIRAC/Setup")

pr = ProductionRequest()

stepsList = ["{{p1Step}}"]
stepsList.append("{{p2Step}}")
stepsList.append("{{p3Step}}")
stepsList.append("{{p4Step}}")
stepsList.append("{{p5Step}}")
stepsList.append("{{p6Step}}")
stepsList.append("{{p7Step}}")
stepsList.append("{{p8Step}}")
stepsList.append("{{p9Step}}")
stepsList.append("{{p10Step}}")
stepsList.append("{{p11Step}}")
stepsList.append("{{p12Step}}")
stepsList.append("{{p13Step}}")
stepsList.append("{{p14Step}}")
stepsList.append("{{p15Step}}")
stepsList.append("{{p16Step}}")
stepsList.append("{{p17Step}}")
stepsList.append("{{p18Step}}")
stepsList.append("{{p19Step}}")
stepsList.append("{{p20Step}}")
pr.stepsList = stepsList

###########################################
# Configurable and fixed parameters
###########################################

pr.appendName = "{{WorkflowAppendName#GENERAL: Workflow string to append to production name#1}}"

w = "{{w#----->WORKFLOW: choose what to do#}}"
pr.prodsTypeList = "{{r#---List here which prods you want to create#WGProduction}}".split(",")
p = "{{p#--now map steps to productions#}}"
p1 = "{{p1#-Production 1 steps (e.g. 1,2,3 = first, second, third)#1,2,3}}"
if p1:
    p1 = p1.split(",")
    p1 = [int(x) for x in p1]
    pr.stepsInProds.append(p1)
p2 = "{{p2#-Production 2 steps (e.g. 4,5)#4,5}}"
if p2:
    p2 = p2.split(",")
    p2 = [int(x) for x in p2]
    pr.stepsInProds.append(p2)
else:
    pr.stepsInProds.append([])

localTestFlag = ast.literal_eval("{{localTestFlag#GENERAL: Set True for local test#False}}")
validationFlag = ast.literal_eval("{{validationFlag#GENERAL: Set True for validation prod#False}}")

# workflow params for all productions
pr.startRun = int("{{startRun#GENERAL: run start, to set the start run#0}}")
pr.endRun = int("{{endRun#GENERAL: run end, to set the end of the range#0}}")
pr.runsList = "{{runsList#GENERAL: discrete list of run numbers (do not mix with start/endrun)#}}"
pr.targets = ["{{WorkflowDestination#GENERAL: Workflow destination site e.g. LCG.CERN.cern#}}"] * len(pr.prodsTypeList)
extraOptions = "{{extraOptions#GENERAL: extra options as python dict stepNumber:options#}}"
if extraOptions:
    pr.extraOptions = ast.literal_eval(extraOptions)
pr.derivedProduction = int("{{AncestorProd#GENERAL: ancestor prod to be derived#0}}")
pr.previousProdID = int("{{previousProdID#GENERAL: previous prod ID (for BK query)#0}}")
modulesList = "{{modulesList#GENERAL: custom modules list#}}"
enablePopularityReport = ast.literal_eval("{{popularityReport#GENERAL: enable popularity report#True}}")
pr.visibility = "{{visibilityFlag#GENERAL: visibility flag for input BK Query#Yes}}"

# p1 params
p1Plugin = "{{p1PluginType#PROD-P1: production plugin name#LHCbStandard}}"
p1Priority = int("{{p1Priority#PROD-P1: priority#2}}")
p1CPU = "{{p1MaxCPUTime#PROD-P1: Max CPU time in secs#1000000}}"
p1GroupSize = "{{p1GroupSize#PROD-P1: Group size or number of files per job#1}}"
p1DataSE = "{{p1DataSE#PROD-P1: Output Data Storage Element#Tier1-DST}}"
try:
    p1DataSESpecial = ast.literal_eval('{{p1DataSESpecial#PROD-P1: Special SEs per file type, e.g. {"T1":"SE1"}#}}')
except SyntaxError:
    p1DataSESpecial = {}
p1Policy = "{{p1Policy#PROD-P1: data policy (download or protocol)#download}}"
p1RemoveInputs = ast.literal_eval("{{p1RemoveInputs#PROD-P1: removeInputs flag#False}}")
p1StepMask = "{{P1StepMask#PROD-P1: step output to save, semicolon separated (default is last)#}}"
p1FileMask = "{{P1FileMask#PROD-P1: file types in output to save, semicolon separated (default is all)#}}"
p1multicoreFlag = "{{P1MulticoreFLag#PROD-P1: multicore flag#True}}"
p1outputMode = "{{P1OutputMode#PROD-P1: output mode#Local}}"
p1eventsRequested = "{{P1EventsRequested#PROD-P1: events requested (-1 = ALL)#-1}}"
p1ancestorDepth = int("{{P1AncestorDepth#PROD-P1: ancestor depth#0}}")

# p2 params
p2Plugin = "{{p2PluginType#PROD-P2: production plugin name#LHCbStandard}}"
p2Priority = int("{{p2Priority#PROD-P2: priority#2}}")
p2CPU = "{{p2MaxCPUTime#PROD-P2: Max CPU time in secs#1000000}}"
p2GroupSize = "{{p2GroupSize#PROD-P2: Group Size#1}}"
p2DataSE = "{{p2DataSE#PROD-P2: Output Data Storage Element#Tier1-DST}}"
try:
    p2DataSESpecial = ast.literal_eval('{{p2DataSESpecial#PROD-P2: Special SEs per file type, e.g. {"T1":"SE1"}#}}')
except SyntaxError:
    p2DataSESpecial = {}
p2Policy = "{{p2Policy#PROD-P2: data policy (download or protocol)#download}}"
p2RemoveInputs = ast.literal_eval("{{p2RemoveInputs#PROD-P2: removeInputs flag#False}}")
p2StepMask = "{{P2StepMask#PROD-P2: step output to save, semicolon separated (default is last#}}"
p2FileMask = "{{P2FileMask#PROD-P2: file types in output to save, semicolon separated (default is all)#}}"
p2multicoreFlag = "{{P2MulticoreFLag#PROD-P2: multicore flag#True}}"
p2outputMode = "{{P2OutputMode#PROD-P2: output mode#Local}}"
p2eventsRequested = "{{P2EventsRequested#PROD-P2: events requested (-1 = ALL)#-1}}"
p2ancestorDepth = int("{{P2AncestorDepth#PROD-P2: ancestor depth#0}}")


numberOfSteps = len(pr.stepsList)
pr.compressionLvl = ["LOW"] * numberOfSteps
pr.compressionLvl[-1] = "HIGH"

vis = {str(i): "N" for i in range(1, numberOfSteps)}
vis[str(numberOfSteps)] = "Y"

pr.outputVisFlag = [vis]
pr.specialOutputVisFlag = [{}]


parentReq = "{{_parent}}"
if not parentReq:
    pr.requestID = "{{ID}}"
else:
    pr.requestID = parentReq

pr.prodGroup = "{{inProPass}}" + "/" + "{{pDsc}}"
# used in case of a test e.g. certification etc.
pr.configName = "{{configName}}"
pr.configVersion = "{{configVersion}}"
pr.outConfigName = pr.configName
# Other parameters from the request page
pr.dqFlag = "{{inDataQualityFlag}}"  # UNCHECKED
pr.dataTakingConditions = "{{simDesc}}"
pr.processingPass = "{{inProPass}}"
if p1[0] == 1 and pr.prodsTypeList[0].lower() != "mcsimulation":
    pr.bkFileType = "{{inFileType}}"
    pr.bkQueries = ["Full"]
elif p1[0] == 1 and pr.prodsTypeList[0].lower() == "mcsimulation":
    pr.bkQueries = [""]
    pr.prodGroup = "{{pDsc}}"
elif p1[0] != 1 and pr.prodsTypeList[0].lower() != "mcsimulation":
    pr.bkQueries = ["fromPreviousProd"]
    if not pr.previousProdID:
        gLogger.error("Please specify an input production")
        DIRACexit(2)
pr.eventType = "{{eventType}}"

if modulesList:
    pr.modulesList = modulesList.replace(" ", "").split(",")

if localTestFlag:
    pr.testFlag = True
    pr.publishFlag = False
    pr.prodsToLaunch = [1]

if validationFlag:
    pr.outConfigName = "validation"

if enablePopularityReport:
    pr.extraModulesList = ["FileUsage"]

inputDataList = []
if not pr.publishFlag:
    testData = "LFN:/lhcb/LHCb/Collision11/CHARMCOMPLETEEVENT.DST/"
    testData += "00012586/0000/00012586_00000706_1.charmcompleteevent.dst"
    inputDataList.append(testData)
    p1Policy = "protocol"
    evtsPerJob = "5"
    bkScriptFlag = True
else:
    bkScriptFlag = False

# In case we want just to test, we publish in the certification/test part of the BKK
if currentSetup == "LHCb-Certification" or pr.testFlag:
    pr.outConfigName = "certification"
    pr.configVersion = "test"

if pr.testFlag:
    pr.startRun = 75336
    pr.endRun = 75340
    p1CPU = "100000"
    pr.events = ["2000"]
    pr.dataTakingConditions = "Beam3500GeV-VeloClosed-MagDown"
    pr.processingPass = "Real Data/Reco12/Stripping17"
    pr.bkFileType = "CHARMCOMPLETEEVENT.DST"
    pr.eventType = "90000000"
    pr.dqFlag = "ALL"

if not p1StepMask:
    p1StepMask = len(pr.stepsInProds[0])
if not p2StepMask:
    try:
        p2StepMask = len(pr.stepsInProds[1])
    except IndexError:
        p2StepMask = ""

pr.outputSEs = [x for x in [p1DataSE, p2DataSE] if x != ""]
pr.specialOutputSEs = [p1DataSESpecial, p2DataSESpecial]
pr.removeInputsFlags = [p1RemoveInputs, p2RemoveInputs][0 : len(pr.prodsTypeList)]
pr.priorities = [p1Priority, p2Priority][0 : len(pr.prodsTypeList)]
pr.cpus = [p1CPU, p2CPU][0 : len(pr.prodsTypeList)]
pr.groupSizes = [p1GroupSize, p2GroupSize][0 : len(pr.prodsTypeList)]
pr.plugins = [p1Plugin, p2Plugin][0 : len(pr.prodsTypeList)]
pr.inputs = [inputDataList, [], []][0 : len(pr.prodsTypeList)]
pr.inputDataPolicies = [p1Policy, p2Policy][0 : len(pr.prodsTypeList)]
pr.outputFileSteps = [p1StepMask, p2StepMask]
pr.outputFileMasks = [p1FileMask, p2FileMask]
pr.multicore = [p1multicoreFlag, p2multicoreFlag]
pr.outputModes = [p1outputMode, p2outputMode]
pr.events = [p1eventsRequested, p2eventsRequested]
pr.ancestorDepths = [p1ancestorDepth, p2ancestorDepth]

pr.buildAndLaunchRequest()



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Provides access to Analysis Productions specific functionality

For more information see :py:mod:`.AnalysisProductionsClient`.
"""
from datetime import datetime
from itertools import chain

from DIRAC import gConfig, gLogger
from DIRAC.Core.Tornado.Server.TornadoService import TornadoService
from DIRAC.Core.Security.Properties import PRODUCTION_MANAGEMENT
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise, convertToReturnValue
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Storage.StorageElement import StorageElement
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.ProductionManagementSystem.DB.AnalysisProductionsDB import AnalysisProductionsDB

optDate = (type(None), datetime)
optString = (type(None), str)
optList = (type(None), list)
optBool = (type(None), bool)

sLog = gLogger.getSubLogger(__name__)
sLog._setOption("tornadoComponent", "ProductionManagement/TornadoAnalysisProduction")


class TornadoAnalysisProductionsHandler(TornadoService):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """initialize handler"""
        cls._db = AnalysisProductionsDB(parentLogger=cls.log)

    types_listAnalyses = [optDate]

    @convertToReturnValue
    def export_listAnalyses(self, at_time):
        """See :meth:`~.AnalysisProductionsClient.listAnalyses`"""
        return self._db.listAnalyses(at_time=at_time)

    types_listAnalyses2 = [optDate]

    @convertToReturnValue
    def export_listAnalyses2(self, at_time):
        """See :meth:`~.AnalysisProductionsClient.listAnalyses2`"""
        return self._db.listAnalyses2(at_time=at_time)

    types_listRequests = [optDate]

    @convertToReturnValue
    def export_listRequests(self):
        """See :meth:`~.AnalysisProductionsClient.listRequests`"""
        return self._db.listRequests()

    types_getOwners = [str, str]

    @convertToReturnValue
    def export_getOwners(self, wg, analysis):
        """See :meth:`~.AnalysisProductionsClient.getOwners`"""
        return self._db.getOwners(wg=wg, analysis=analysis)

    types_setOwners = [str, str, list]

    @convertToReturnValue
    def export_setOwners(self, wg, analysis, owners):
        """See :meth:`~.AnalysisProductionsClient.setOwners`"""
        return self._db.setOwners(wg=wg, analysis=analysis, owners=owners)

    types_getProductions = [
        optString,
        optString,
        optString,
        optString,
        optString,
        bool,
        bool,
        bool,
        optDate,
        optBool,
        optBool,
    ]

    @convertToReturnValue
    def export_getProductions(
        self,
        wg,
        analysis,
        version,
        name,
        state,
        with_lfns,
        with_pfns,
        with_transformations,
        at_time,
        show_archived=False,
        require_has_publication=False,
    ):
        """See :meth:`~.AnalysisProductionsClient.getProductions`"""
        if (analysis or name or version) and wg is None:
            raise TypeError("wg must be specified when providing an analysis name or sample name/version")
        if with_pfns and not with_lfns:
            raise TypeError("with_pfns cannot be used without with_lfns")
        results = self._db.getProductions(
            wg=wg,
            analysis=analysis,
            version=version,
            name=name,
            state=state,
            at_time=at_time,
            show_archived=show_archived,
            require_has_publication=require_has_publication,
        )
        return _queryToResults(results, with_lfns, with_pfns, with_transformations)

    types_getArchivedRequests = [optString, bool, bool, bool]

    @convertToReturnValue
    def export_getArchivedRequests(self, state, with_lfns, with_pfns, with_transformations):
        """See :meth:`~.AnalysisProductionsClient.getArchivedRequests`"""
        if with_pfns and not with_lfns:
            raise TypeError("with_pfns cannot be used without with_lfns")
        results = self._db.getArchivedRequests(state=state)
        return _queryToResults(results, with_lfns, with_pfns, with_transformations)

    types_getTags = [optString, optString, optDate]

    @convertToReturnValue
    def export_getTags(self, wg, analysis, at_time):
        """See :meth:`~.AnalysisProductionsClient.getTags`"""
        return self._db.getTags(wg, analysis, at_time=at_time)

    types_getKnownAutoTags = []

    @convertToReturnValue
    def export_getKnownAutoTags(self):
        """See :meth:`~.AnalysisProductionsClient.getKnownAutoTags`"""
        return list(self._db.getKnownAutoTags())

    types_registerTransformations = [dict]

    @convertToReturnValue
    def export_registerTransformations(self, transforms):
        """See :meth:`~.AnalysisProductionsClient.registerTransformations`"""
        transforms = {int(k): v for k, v in transforms.items()}
        return self._db.registerTransformations(transforms)

    types_deregisterTransformations = [dict]

    @convertToReturnValue
    def export_deregisterTransformations(self, tIDs):
        """See :meth:`~.AnalysisProductionsClient.deregisterTransformations`"""
        tIDs = {int(k): v for k, v in tIDs.items()}
        return self._db.deregisterTransformations(tIDs)

    types_registerRequests = [list]

    @convertToReturnValue
    def export_registerRequests(self, requests):
        """See :meth:`~.AnalysisProductionsClient.registerRequests`"""
        results = self._db.registerRequests(requests)
        return _queryToResults(results, with_lfns=False, with_pfns=False, with_transformations=False)

    types_addRequestsToAnalysis = [str, str, list[tuple[int, str]]]

    @convertToReturnValue
    def export_addRequestsToAnalysis(self, wg, analysis, requests):
        """See :meth:`~.AnalysisProductionsClient.addRequestsToAnalysis`"""
        return self._db.addRequestsToAnalysis(wg, analysis, requests)

    types_archiveSamples = [list]

    @convertToReturnValue
    def export_archiveSamples(self, sample_ids):
        """See :meth:`~.AnalysisProductionsClient.archiveSamples`"""
        return self._db.archiveSamples(sample_ids)

    types_archiveSamplesAtSpecificTime = [list, datetime]

    @convertToReturnValue
    def export_archiveSamplesAtSpecificTime(self, sample_ids, archive_time):
        """See :meth:`~.AnalysisProductionsClient.archiveSamplesAtSpecificTime`"""
        return self._db.archiveSamplesAtSpecificTime(sample_ids, archive_time)

    types_setState = [dict]

    @convertToReturnValue
    def export_setState(self, newState):
        """See :meth:`~.AnalysisProductionsClient.setState`"""
        newState = {(int(k), ft): vv for k, v in newState.items() for ft, vv in v.items()}
        return self._db.setState(newState)

    types_setTags = [dict, dict]

    @convertToReturnValue
    def export_setTags(self, oldTags, newTags):
        """See :meth:`~.AnalysisProductionsClient.setTags`"""
        return self._db.setTags(oldTags, newTags)

    types_delayHousekeepingInteractionDue = [list, datetime]

    @convertToReturnValue
    def export_delayHousekeepingInteractionDue(self, samples, next_interaction_due):
        """See :meth:`~.AnalysisProductionsClient.delayHousekeepingInteractionDue`"""
        return self._db.delayHousekeepingInteractionDue(samples, next_interaction_due)

    types_getHousekeepingInteractionDueNow = []

    @convertToReturnValue
    def export_getHousekeepingInteractionDueNow(self):
        """See :meth:`~.AnalysisProductionsClient.getHousekeepingInteractionDueNow`"""
        return self._db.getHousekeepingInteractionDueNow()

    types_addPublication = [list, str]

    @convertToReturnValue
    def export_addPublication(self, samples, number):
        """See :meth:`~.AnalysisProductionsClient.addPublication`"""
        return self._db.addPublication(samples, number)

    types_getPublications = [optList]

    @convertToReturnValue
    def export_getPublications(self, sample_ids):
        """See :meth:`~.AnalysisProductionsClient.getPublications`"""
        return self._db.getPublications(sample_ids)


def _queryToResults(results, with_lfns, with_pfns, with_transformations):
    """Convert a query of Request or AP objects to a JSON compatible object"""
    p2tID = {
        (result["request_id"], result["filetype"]): {tInfo["id"]: tInfo["used"] for tInfo in result["transformations"]}
        for result in results
    }

    if with_lfns:
        lfns = _getOutputLFNs(p2tID)
        if with_pfns:
            replicas = _getReplicas(list(chain(*lfns.values())))

        for result in results:
            result["lfns"] = {} if with_pfns else []
            result["total_bytes"] = 0
            if with_pfns:
                result["available_bytes"] = 0
            for tInfo in result["transformations"]:
                for lfn, lfnMeta in lfns.get(tInfo["id"], {}).items():
                    result["total_bytes"] += lfnMeta["FileSize"]
                    if not with_pfns:
                        result["lfns"].append(lfn)
                    # TODO: Remove when https://github.com/pylint-dev/pylint/issues/9628 is fixed
                    elif lfn in replicas["Failed"]:  # pylint: disable=possibly-used-before-assignment
                        result["lfns"][lfn] = []
                        sLog.warn(
                            "Failed to get replicas for Analysis Production LFN",
                            "%s in rID=%s tID=%s: %s"
                            % (lfn, result["request_id"], tInfo["id"], replicas["Failed"][lfn]),
                        )
                    else:
                        result["lfns"][lfn] = [
                            pfn
                            for pfn in replicas["Successful"][lfn].values()
                            # Ignore entries with the value True as these are at unsupported SEs
                            if isinstance(pfn, str)
                        ]
                        if result["lfns"][lfn]:
                            result["available_bytes"] += lfnMeta["FileSize"]

    if with_transformations:
        tIDs = list(set(chain(*p2tID.values())))
        if tIDs:
            tClient = TransformationClient()
            retVal = tClient.getTransformations(
                condDict={"TransformationID": tIDs},
                limit=10000,
                columns=["TransformationID", "Status", "Type"],
            )
            extraTransInfos = {t["TransformationID"]: t for t in returnValueOrRaise(retVal)}
            # Query any transformation save for types without well defined input queries
            inputQueries = returnValueOrRaise(
                tClient.getBookkeepingQueries(
                    [tID for tID in tIDs if extraTransInfos[tID]["Type"] not in ["DerriviedAnalysisProduction"]]
                )
            )
            for tInfo in chain(*(r["transformations"] for r in results)):
                tInfo["status"] = extraTransInfos[tInfo["id"]]["Status"]
                tInfo["input_query"] = inputQueries.get(tInfo["id"], {})
    else:
        for result in results:
            del result["transformations"]

    return results


def _getOutputLFNs(pID2tID):
    sLog.info("Getting output LFNs")
    keys = [(tID, ft) for (_, ft), tIDs in pID2tID.items() for tID, used in tIDs.items() if used]
    lfns = {}
    if not keys:
        return lfns
    retVal = BookkeepingClient().getProductionFilesBulk(
        *map(list, zip(*keys, strict=True)),
        "ALL",
    )
    for tID, lfnMetadata in returnValueOrRaise(retVal).items():
        lfns[int(tID)] = {lfn: meta for lfn, meta in lfnMetadata.items() if meta["GotReplica"].lower().startswith("y")}
    return lfns


def _getReplicas(lfns):
    sLog.info("Getting replicas for", f"{len(lfns)} LFNs")
    replicas = returnValueOrRaise(DataManager().getReplicas(lfns, diskOnly=True, getUrl=False))
    for se in gConfig.getValue("/Resources/Sites/LCG/LCG.CERN.cern/SE", []):
        seLFNs = {lfn for lfn, replicaInfo in replicas["Successful"].items() if se in replicaInfo}
        if not seLFNs:
            continue
        sLog.info("Getting PFNs for", f"{len(seLFNs)} LFNs at {se}")
        result = returnValueOrRaise(StorageElement(se).getURL(list(seLFNs), protocol="root"))
        if result["Failed"]:
            raise NotImplementedError(result["Failed"])
        for lfn, pfn in result["Successful"].items():
            replicas["Successful"][lfn][se] = pfn
    return replicas



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Retrieve from the Bookkeeping runs from a given date range."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from DIRAC import gLogger, exit as DIRACexit

    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] ... Start [End]",
                "Arguments:",
                "  Start:    Start date (Format: YYYY-MM-DD) (mandatory)",
                "  End:      End date (Format: YYYY-MM-DD). Default is Start",
            ]
        )
    )
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    start = ""
    end = ""
    if len(args) < 2:
        Script.showHelp(exitCode=1)

    if len(args) == 2:
        end = args[1]
    start = args[0]

    in_dict = {}
    in_dict["StartDate"] = start
    in_dict["EndDate"] = end if end else start

    res = BookkeepingClient().getRunsForAGivenPeriod(in_dict)
    if not res["OK"]:
        gLogger.error(f"Failed to retrieve runs: {res['Message']}")
        DIRACexit(1)

    if not res["Value"]["Runs"]:
        gLogger.notice("No runs found for the date range", (start, end))
    else:
        gLogger.notice("Runs:", res["Value"]["Runs"])
        if "ProcessedRuns" in res["Value"]:
            gLogger.notice("Processed runs:", res["Value"]["ProcessedRuns"])
        if "NotProcessedRuns" in res["Value"]:
            gLogger.notice("Not processed runs:", res["Value"]["NotProcessedRuns"])


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set of functions used by the DMS scripts."""
import os
import datetime

from DIRAC import gLogger, S_OK, exit as diracExit
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Base.Script import Script
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.DataManagementSystem.Client.DMScript import printDMResult, ProgressBar
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery, parseRuns, BadRunRange, getProcessingPasses
from LHCbDIRAC.BookkeepingSystem.Client.LHCB_BKKDBClient import LHCB_BKKDBClient

bkClient = BookkeepingClient()
jobEventInputStat = {}


def executeFileMetadata(dmScript):
    """Get a list of LFNs, their BK metadata and print it out.

    :param dmScript: instance containing options (LFNs mandatory)
    :type dmScript: DMScript class
    """
    full = False
    switches = Script.getUnprocessedSwitches()
    for switch in switches:
        if switch[0] == "Full":
            full = True

    args = Script.getPositionalArgs()
    for lfn in args:
        dmScript.setLFNsFromFile(lfn)
    lfnList = sorted(dmScript.getOption("LFNs", []))

    if not lfnList:
        Script.showHelp()

    res = bkClient.getFileMetadata(lfnList)
    if not res["OK"]:
        gLogger.error("ERROR: Failed to get file metadata:", res["Message"])
        diracExit(2)

    exitCode = 0
    lenName = max(len(lfn) for lfn in lfnList) + 2
    lenGUID = 38
    lenItem = 15
    sep = ""

    if not full:
        # Header: justification must match those in the loop!
        gLogger.notice(
            "%s %s %s %s %s %s %s"
            % (
                "FileName".ljust(lenName),
                "Size".ljust(10),
                "GUID".ljust(lenGUID),
                "Replica".ljust(8),
                "DataQuality".ljust(12),
                "RunNumber".ljust(10),
                "#events".ljust(10),
            )
        )
    lfnMetadata = res["Value"]["Successful"]
    for lfn in lfnMetadata:
        lfnMetaDict = lfnMetadata[lfn]
        if full:
            gLogger.notice(f"{sep}{'FileName'.ljust(lenItem)} {lfn}")
            sep = "\n"
            for item in sorted(lfnMetaDict):
                gLogger.notice(f"{item.ljust(lenItem)} {lfnMetaDict[item]}")
        else:
            size = lfnMetaDict["FileSize"]
            guid = lfnMetaDict["GUID"]
            gotReplica = lfnMetaDict["GotReplica"]
            dq = lfnMetaDict.get("DataqualityFlag")
            run = lfnMetaDict["RunNumber"]
            evtStat = lfnMetaDict["EventStat"]
            if not gotReplica:
                gotReplica = "No"
            gLogger.notice(
                "%s %s %s %s %s %s %s"
                % (
                    lfn.ljust(lenName),
                    str(size).ljust(10),
                    guid.ljust(lenGUID),
                    gotReplica.ljust(8),
                    dq.ljust(12),
                    str(run).ljust(10),
                    str(evtStat).ljust(10),
                )
            )
    failed = res["Value"].get("Failed", [])
    if failed:
        gLogger.notice("\n")
        for lfn in failed:
            gLogger.error("File does not exist in the Bookkeeping.", lfn)
            exitCode = 2

    diracExit(exitCode)


def __buildPath(bkDict):
    """Build a BK path from the BK dictionary."""
    return os.path.join(
        "/" + bkDict["ConfigName"],
        bkDict["ConfigVersion"],
        bkDict["ConditionDescription"],
        bkDict["ProcessingPass"][1:].replace("Real Data", "RealData"),
        str(bkDict["EventType"]),
        bkDict["FileType"],
    ) + (" (Invisible)" if bkDict["VisibilityFlag"] == "N" else "")


def executeFilePath(dmScript):
    """Gets a list of LFNs and extracts their BK paths or other metadata Then
    print the result or group LFNs by path or metadata.

    :param dmScript: instance containing options (LFNs mandatory)
    :type dmScript: DMScript class
    """
    full = False
    groupBy = False
    summary = False
    printList = False
    ignoreFileType = False
    switches = Script.getUnprocessedSwitches()
    for switch in switches:
        if switch[0] == "Full":
            full = True
        elif switch[0] == "GroupByPath":
            groupBy = "Path"
        elif switch[0] == "GroupByProduction":
            groupBy = "Production"
        elif switch[0] == "GroupBy":
            groupBy = switch[1]
        elif switch[0] == "Summary":
            summary = True
        elif switch[0] == "List":
            printList = True
        elif switch[0] == "IgnoreFileType":
            ignoreFileType = True
    if summary and not groupBy:
        groupBy = "Path"

    args = Script.getPositionalArgs()
    for lfn in args:
        dmScript.setLFNsFromFile(lfn)
    lfnList = sorted(dmScript.getOption("LFNs", []))

    if not lfnList:
        lfnList = sorted(dmScript.getOption("Directory", []))
        if not lfnList:
            Script.showHelp()

    dirMetadata = (
        "Production",
        "ConfigName",
        "ConditionDescription",
        "EventType",
        "FileType",
        "ConfigVersion",
        "ProcessingPass",
        "Path",
    )
    fileMetadata = ("EventType", "FileType", "RunNumber", "JobId", "DataqualityFlag", "GotReplica", "VisibilityFlag")
    if groupBy and groupBy not in dirMetadata and groupBy not in fileMetadata:
        for meta in dirMetadata + fileMetadata:
            if groupBy.lower() == meta.lower():
                groupBy = meta
                break
    if groupBy and groupBy not in dirMetadata:
        if groupBy not in fileMetadata:
            gLogger.notice("Invalid metata item", groupBy)
            gLogger.notice("Directory metadata:", ", ".join(dirMetadata))
            gLogger.notice("File metadata:", ", ".join(fileMetadata))
            diracExit(1)
        paths = {"Successful": {}, "Failed": []}
        chunkSize = 1000
        progressBar = ProgressBar(
            len(lfnList), title=f"Getting {groupBy} from BK for {len(lfnList)} files", chunk=chunkSize
        )
        for lfnChunk in breakListIntoChunks(lfnList, chunkSize):
            res = bkClient.getFileMetadata(lfnChunk)
            progressBar.loop()
            if res["OK"]:
                for lfn, metadata in res["Value"]["Successful"].items():  # can be an iterator
                    group = metadata.get(groupBy)
                    paths["Successful"].setdefault(f"{groupBy} {group}", set()).add(lfn)
                    lfnChunk.remove(lfn)
                paths["Failed"].extend(lfnChunk)
        progressBar.endLoop()
        if summary:
            pathSummary = {"Successful": {}}
            for groupStr in paths["Successful"]:
                pathSummary["Successful"][groupStr] = f"{len(paths['Successful'][groupStr])} files"
            if paths["Failed"]:
                pathSummary["Failed"] = len(paths["Failed"])
            res = S_OK(pathSummary)
        else:
            res = S_OK(paths)
    else:
        directories = {}
        for lfn in lfnList:
            dirName = os.path.dirname(lfn)
            if not dirName:
                continue
            if "/RAW/" not in dirName:
                # If it is is RAW, we already had the correct path.
                # for example: os.path.dirname('/lhcb/data/2016/RAW/TURBO/LHCb/COLLISION16/176059/176059_0000003101.raw')
                #    returns /lhcb/data/2016/RAW/TURBO/LHCb/COLLISION16/176059
                tail = os.path.basename(dirName)
                # Eliminate the tailing '/0000'
                if len(tail) == 4 and tail.isdigit():
                    dirName = os.path.dirname(dirName)
            directories.setdefault(dirName, []).append(lfn)

        chunkSize = 2
        progressBar = ProgressBar(
            len(directories), title=f"Getting metadata for {len(directories)} directories", chunk=chunkSize
        )
        success = {}
        failed = set()
        for dirChunk in breakListIntoChunks(directories, chunkSize):
            progressBar.loop()
            res = bkClient.getDirectoryMetadata(dirChunk)
            if not res["OK"]:
                progressBar.endLoop(message="Error getting directory metadata")
                printDMResult(res)
                diracExit(1)
            success.update(res["Value"]["Successful"])
            failed.update(res["Value"]["Failed"])
        progressBar.endLoop()

        res = S_OK({"Successful": success, "Failed": failed})
        paths = {"Successful": {}, "Failed": {}}
        for dirName in success:
            # If there is more than on result, they differ by the visibility
            bestMetadata = None  # to make pylint happy
            for bestMetadata in success[dirName]:
                # It happens that some files are visible and others invisible,
                #   therefore select the metadata that is visible if any
                if bestMetadata["VisibilityFlag"] == "Y":
                    break
            if full:
                success[dirName] = bestMetadata
            else:
                bkDict = bestMetadata
                if ignoreFileType:
                    bkDict["FileType"] = ""
                bkDict["Path"] = __buildPath(bkDict)
                if groupBy in bkDict:
                    if groupBy != "Path":
                        prStr = f"{groupBy} {bkDict[groupBy]}"
                    else:
                        prStr = bkDict[groupBy]
                    paths["Successful"].setdefault(prStr, set()).update(directories[dirName])
                elif groupBy:
                    gLogger.notice(f"Invalid metadata item: {groupBy}")
                    gLogger.notice(f"Available are: {str(list(bkDict))}")
                    diracExit(1)
                else:
                    success[dirName] = bkDict["Path"]

        if groupBy:
            if summary:
                pathSummary = {"Successful": {}, "Failed": {}}
                for path in paths["Successful"]:
                    nfiles = len(paths["Successful"][path])
                    if "Invisible" in path:
                        path = path.split()[0]
                        inv = " (Invisible)"
                    else:
                        inv = ""
                    pathSummary["Successful"][path] = "%d files%s" % (nfiles, inv)
                if failed:
                    pathSummary["Failed"] = {
                        path: f"Directory not in BK ({len(directories[path])} files)" for path in failed
                    }
                else:
                    pathSummary.pop("Failed")
                res = S_OK(pathSummary)
            else:
                for dirName in failed:
                    paths["Failed"].update(dict.fromkeys(directories[dirName], "Directory not in BK"))
                res = S_OK(paths)

    printDMResult(res, empty="None", script="dirac-bookkeeping-file-path")
    if printList:
        gLogger.notice(f"\nList of {groupBy} values")
        gLogger.notice(",".join(sorted(item.replace(f"{groupBy} ", "") for item in res["Value"]["Successful"])))


def _updateFileLumi(fileDict, retries=5):
    """Update the luminosity of a list of files in the BK.

    :param dict fileDict: {lfn:luminosity}

    :return: bool reporting error
    """
    error = False
    progressBar = ProgressBar(len(fileDict), title="Updating luminosity", step=10)
    for lfn in fileDict:
        progressBar.loop()
        # retry 5 times
        for i in range(retries - 1, -1, -1):
            res = bkClient.updateFileMetaData(lfn, {"Luminosity": fileDict[lfn]})
            if res["OK"]:
                break
            elif i == 0:
                error = True
                gLogger.error("Error setting Luminosity", res["Message"])
    progressBar.endLoop()
    return error


def _updateDescendantsLumi(parentLumi, doIt=False, force=False):
    """Get file descendants and update their luminosity if necessary (if doIt ==
    True) This function does it recursively to all descendants.

    :param dict parentLumi: {lfn:lumi} for parent files
    :param bool doit: execute the operation if True, else only print out information
    :param bool force: update lumi even if OK (useful if further descendants may not be OK)

    :return: bool indicating error
    """
    if not parentLumi:
        return None
    # Get descendants:
    error = False
    res = bkClient.getFileDescendents(list(parentLumi), depth=1, checkreplica=False)
    if not res["OK"]:
        gLogger.error("Error getting descendants", res["Message"])
        return True
    success = res["Value"]["WithMetadata"]
    descLumi = {}
    fileTypes = {}
    fileLumi = {}
    for lfn in success:
        for desc in success[lfn]:
            fileType = success[lfn][desc]["FileType"]
            if fileType not in ("LOG",) and "HIST" not in fileType:
                descLumi.setdefault(desc, 0.0)
                descLumi[desc] += parentLumi[lfn]
                fileTypes[desc] = fileType
                fileLumi[desc] = success[lfn][desc]["Luminosity"]
    if not descLumi:
        return None

    prStr = "Updating" if doIt else "Would update"
    nDesc = len(descLumi)
    saveLumi = descLumi.copy()
    for lfn in fileLumi:
        if abs(fileLumi[lfn] - descLumi[lfn]) < 1:
            descLumi.pop(lfn, None)
    if descLumi:
        gLogger.notice(
            "%s lumi of %d descendants out of %d (file types: %s) of %d files"
            % (prStr, len(descLumi), nDesc, ",".join(sorted(set(fileTypes.values()))), len(parentLumi))
        )
    else:
        gLogger.notice(
            "All %d descendants (file types: %s) of %d files are OK"
            % (nDesc, ",".join(sorted(set(fileTypes.values()))), len(parentLumi))
        )
        if not force:
            return None
    if doIt:
        error = _updateFileLumi(descLumi)
    result = _updateDescendantsLumi(descLumi if not force else saveLumi, doIt=doIt, force=force)
    return error or bool(result)


def _updateRunLumi(run, evtType, fileInfo, doIt=False, force=False):
    """
    Updates the files luminosity from the run nformation and the files statistics
    run : run number (int)
    evtType: event type (int)
    fileInfo: list of tuples containing run files information for that event type [(lfn, nbevts, lumi), ...]
    """
    res = bkClient.getRunInformations(run)
    if not res["OK"]:
        gLogger.error("Error from BK getting run information", res["Message"])
        return
    info = res["Value"]
    runLumi = info["TotalLuminosity"]
    runEvts = dict(zip(info["Stream"], info["Number of events"]))[evtType]
    filesLumi = sum(lumi for _lfn, _evts, lumi in fileInfo)
    # Check luminosity
    error = False
    if not runEvts:
        gLogger.notice("Event number is zero for run %d, cannot update" % run)
    elif abs(runLumi - filesLumi) > 1:
        prStr = "Updating" if doIt else "Would update"
        gLogger.notice(
            "%s %d files as run %d and files lumi don't match: runLumi %.1f, filesLumi %.1f"
            % (prStr, len(fileInfo), run, runLumi, filesLumi)
        )
        fileDict = {}
        for info in fileInfo:
            # Split the luminosity according to nb of events
            info[2] = float(runLumi) * info[1] / runEvts
            fileDict[info[0]] = info[2]
        if doIt:
            error = _updateFileLumi(fileDict)
    else:
        gLogger.notice("Run %d: %d RAW files are OK" % (run, len(fileInfo)))

    # Now update descendants
    fileLumi = {lfn: lumi for lfn, _evts, lumi in fileInfo}
    result = _updateDescendantsLumi(fileLumi, doIt=doIt, force=force)
    return error or result


def executeFixLuminosity(dmScript):
    """Checks the luminosity of files in a BK query against the run luminosity If
    requested, it fixes recursively the luminosity of the files and all
    descendants.

    :param dmScript: instance containing options (BKQuery mandatory)
    :type dmScript: DMScript class
    """
    doIt = False
    force = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "DoIt":
            doIt = True
        elif switch[0] == "Force":
            force = True

    bkQuery = dmScript.getBKQuery()
    if not bkQuery:
        gLogger.error("No BK query given...")
        diracExit(1)

    fileType = bkQuery.getFileTypeList()
    if fileType != ["RAW"]:
        gLogger.notice("This script only works from RAW files")
        diracExit(2)

    bkQueryDict = bkQuery.getQueryDict()
    evtTypes = bkQuery.getEventTypeList()

    runChecked = False
    for evtType in evtTypes:
        gLogger.notice(f"**** Event type {evtType}")
        bkQueryDict["EventType"] = evtType
        res = bkClient.getFilesWithMetadata(bkQueryDict)
        if not res["OK"]:
            gLogger.fatal("Error getting BK files", res["Message"])
        parameterNames = res["Value"]["ParameterNames"]
        info = res["Value"]["Records"]
        runFiles = {}
        for item in info:
            metadata = dict(zip(parameterNames, item))
            run = metadata["RunNumber"]
            lfn = metadata["FileName"]
            lumi = metadata["Luminosity"]
            evts = metadata["EventStat"]
            runFiles.setdefault(run, []).append([lfn, evts, lumi])
        if not runChecked:
            res = bkClient.getRunStatus(list(runFiles))
            if not res["OK"]:
                gLogger.fatal("Error getting run status", res["Message"])
                diracExit(3)
            runNotFinished = sorted(
                str(run) for run in res["Value"]["Successful"] if res["Value"]["Successful"][run]["Finished"] != "Y"
            )
            if runNotFinished:
                gLogger.notice(f"Found {len(runNotFinished)} runs that are not Finished: {','.join(runNotFinished)}")
            runFinished = sorted(
                run for run in res["Value"]["Successful"] if res["Value"]["Successful"][run]["Finished"] == "Y"
            )
            if runFinished:
                gLogger.notice(f"Found {len(runFinished)} runs that are Finished")
            else:
                gLogger.notice("No Finished run found")
                diracExit(0)
        for run in runFinished:
            result = _updateRunLumi(run, int(evtType), runFiles[run], doIt=doIt, force=force)
            if doIt:
                if result is not None:
                    gLogger.notice(f"Update done {'with errors' if result else 'successfully'}")


def executeFileAncestors(dmScript, level=1):
    """Gets a list of LFNs and obtains from BK the list of ancestors at a certain
    depth and/or for a certain Production.

    :param dmScript: instance containing options (LFNs mandatory)
    :type dmScript: DMScript class
    :param level: level to which search for ancestors
    :type level: int
    """
    full = False
    checkreplica = True
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "Full":
            full = True
        elif switch[0] == "All":
            checkreplica = False
        elif switch[0] == "Depth":
            try:
                level = int(switch[1])
            except ValueError as e:
                gLogger.exception("Invalid value for --Depth:", switch[1], lException=e)

    args = Script.getPositionalArgs()

    try:
        level = int(args[-1])
        args.pop()
    except (ValueError, IndexError):
        pass

    for lfn in args:
        dmScript.setLFNsFromFile(lfn)
    lfnList = dmScript.getOption("LFNs", [])

    chunkSize = 50
    progressBar = ProgressBar(
        len(lfnList), chunk=chunkSize, title="Getting ancestors for %d files (depth %d)" % (len(lfnList), level)
    )
    fullResult = S_OK({})
    for lfnChunk in breakListIntoChunks(lfnList, 50):
        progressBar.loop()
        result = bkClient.getFileAncestors(lfnChunk, level, replica=checkreplica)

        if result["OK"]:
            for lfn in set(lfnChunk) - set(result["Value"]["Successful"]) - set(result["Value"]["Failed"]):
                fullResult["Value"].setdefault("Failed", {})[lfn] = "No ancestor found"
            if full:
                fullResult["Value"].setdefault("WithMetadata", {}).update(result["Value"]["WithMetadata"])
            else:
                okResult = result["Value"]["WithMetadata"]
                for lfn in okResult:
                    fullResult["Value"].setdefault("Successful", {})[lfn] = {
                        desc: f"Replica-{meta['GotReplica']}" for desc, meta in okResult[lfn].items()
                    }  # can be an iterator
            failed = result["Value"]["Failed"]
            if isinstance(failed, list):
                failed = dict.fromkeys(failed, "Not found")
            fullResult["Value"].setdefault("Failed", {}).update(failed)
        else:
            fullResult = result
            break
    progressBar.endLoop()

    diracExit(printDMResult(fullResult, empty="None", script="dirac-bookkeeping-get-file-ancestors"))


def executeFileDescendants(dmScript, level=1):
    """Gets a list of LFNs and obtains from BK the list of descendants at a
    certain depth and/or for a certain Production.

    :param dmScript: instance containing options (LFNs mandatory)
    :type dmScript: DMScript class
    :param level: level to which search for descendants
    :type level: int
    """
    checkreplica = True
    prod = 0
    full = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "All":
            checkreplica = False
        elif switch[0] == "Depth":
            try:
                level = int(switch[1])
            except ValueError as e:
                gLogger.exception("Invalid value for --Depth: %s", switch[1], lException=e)
        elif switch[0] == "Production":
            try:
                prod = int(switch[1])
            except ValueError as e:
                gLogger.exception("Invalid production", switch[1], lException=e)
                prod = 0
        elif switch[0] == "Full":
            full = True

    args = Script.getPositionalArgs()

    try:
        level = int(args[-1])
        args.pop()
    except (ValueError, IndexError):
        pass

    for lfn in args:
        dmScript.setLFNsFromFile(lfn)
    lfnList = dmScript.getOption("LFNs", [])

    chunkSize = int(20 / level)
    progressBar = ProgressBar(
        len(lfnList),
        chunk=chunkSize,
        title="Getting descendants for %d files (depth %d)" % (len(lfnList), level)
        + (" for production %d" % prod if prod else ""),
    )
    fullResult = S_OK({})
    for lfnChunk in breakListIntoChunks(lfnList, 50):
        progressBar.loop()
        result = bkClient.getFileDescendents(lfnChunk, depth=level, production=prod, checkreplica=checkreplica)
        if result["OK"]:
            noDescendants = (
                set(lfnChunk)
                - set(result["Value"]["Successful"])
                - set(result["Value"]["Failed"])
                - set(result["Value"]["NotProcessed"])
            )
            if noDescendants:
                fullResult["Value"].setdefault("NoDescendants", []).extend(sorted(noDescendants))
            if full:
                fullResult["Value"].setdefault("WithMetadata", {}).update(result["Value"]["WithMetadata"])
            else:
                okResult = result["Value"]["WithMetadata"]
                for lfn in okResult:
                    fullResult["Value"].setdefault("Successful", {})[lfn] = {
                        desc: f"Replica-{meta['GotReplica']}" for desc, meta in okResult[lfn].items()
                    }  # can be an iterator
            failed = result["Value"]["Failed"]
            if isinstance(failed, list):
                failed = dict.fromkeys(failed, "Unknown error")
            fullResult["Value"].setdefault("Failed", {}).update(failed)
            fullResult["Value"].setdefault("NotProcessed", []).extend(result["Value"]["NotProcessed"])
        else:
            fullResult = result
            break
    progressBar.endLoop()

    diracExit(printDMResult(fullResult, empty="None", script="dirac-bookkeeping-get-file-descendants"))


def executeGetFiles(dmScript, maxFiles=20):
    """Get files given a BK query.

    :param dmScript: instance containing options (BKQuery mandatory)
    :type dmScript: DMScript class
    :param maxFiles: maximum number of files to get
    :type maxFiles: int
    """
    output = None
    nMax = None
    bkFile = None
    optionsFile = None
    for switch, val in Script.getUnprocessedSwitches():
        if switch == "Output":
            output = val
        elif switch == "File":
            bkFile = val
        elif switch == "OptionsFile":
            optionsFile = val
        elif switch == "Term":
            bkFile = "/dev/stdin"
        elif switch == "MaxFiles":
            try:
                nMax = int(val)
            except ValueError:
                gLogger.error("Invalid integer", val)
                diracExit(2)

    bkQuery = dmScript.getBKQuery()
    bkQueries = [bkQuery] if bkQuery else []
    if bkFile and os.path.exists(bkFile):
        with open(bkFile) as fd:
            bkQueries += [BKQuery(ll.strip().split()[0]) for ll in fd.readlines()]

    if not bkQueries:
        gLogger.notice("No BK query given, use --BKQuery <bkPath> or <localFile>")
        diracExit(1)

    fileDict = {}
    for bkQuery in bkQueries:
        gLogger.notice("Using BKQuery:", bkQuery)

        useFilesWithMetadata = False
        if useFilesWithMetadata:
            res = bkClient.getFilesWithMetadata(bkQuery.getQueryDict())
            if not res["OK"]:
                gLogger.error("ERROR getting the files", res["Message"])
                diracExit(1)
            parameters = res["Value"]["ParameterNames"]
            for record in res["Value"]["Records"]:
                dd = dict(zip(parameters, record))
                lfn = dd.pop("FileName")
                fileDict[lfn] = dd

        else:
            lfns = bkQuery.getLFNs(printSEUsage=False, printOutput=False)

            for lfnChunk in breakListIntoChunks(lfns, 1000):
                res = bkClient.getFileMetadata(lfnChunk)
                if not res["OK"]:
                    gLogger.error("ERROR: failed to get metadata:", res["Message"])
                    diracExit(1)
                fileDict.update(res["Value"]["Successful"])

    if not fileDict:
        gLogger.notice("No files found for BK query")
        diracExit(0)

    # Now print out
    nFiles = len(fileDict)
    if output:
        fd = open(output, "w")
        if not nMax:
            nMax = maxFiles
    else:
        if not nMax:
            nMax = nFiles
    gLogger.notice("%d files found" % nFiles, "(showing only first %d files):" % nMax if nFiles > nMax else ":")
    outputStr = "{} {} {} {} {}".format(
        "FileName".ljust(100),
        "Size".ljust(10),
        "GUID".ljust(40),
        "Replica".ljust(8),
        "Visibility".ljust(8),
    )
    gLogger.notice(outputStr)
    nFiles = 0
    for lfn in sorted(fileDict):
        metadata = fileDict[lfn]
        size = metadata["FileSize"]
        guid = metadata["GUID"]
        hasReplica = metadata["GotReplica"]
        visible = metadata.get("VisibilityFlag", "?")
        outputStr = "{} {} {} {} {}".format(
            lfn.ljust(100),
            str(size).ljust(10),
            guid.ljust(40),
            str(hasReplica).ljust(8),
            str(visible).ljust(8),
        )
        if output:
            fd.write(outputStr + "\n")
        if nFiles < nMax:
            nFiles += 1
            gLogger.notice(outputStr)

    if output:
        fd.close()
        gLogger.notice(f"\nList of {len(fileDict)} files saved in", output)

    if optionsFile:
        # Use BK client method to write options file
        if len(bkQueries) == 1:
            dataset = bkQueries[0].getQueryDict()
            # RunNumber screws up step info
            for item in ("RunNumber", "RunStart", "RunEnd"):
                dataset.pop(item, None)
            dataset["fullpath"] = bkQueries[0].getPath()
        else:
            dataset = None
        LHCB_BKKDBClient().writeJobOptions(fileDict, optionsFile=optionsFile, dataset=dataset)
        gLogger.notice(f"\n{len(fileDict)} files in options file {optionsFile}")


def executeFileSisters(dmScript, level=1):
    """Gets a list of files and extract from BK the sisters (i.e. files with same
    parent in the same production.

    :param dmScript: instance containing options (LFNs mandatory)
    :type dmScript: DMScript class
    :param level: level to which search for common ancestors
    :type level: int
    """
    checkreplica = True
    prod = 0
    full = False
    sameType = True
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "All":
            checkreplica = False
        elif switch[0] == "Depth":
            try:
                level = int(switch[1])
            except ValueError as e:
                gLogger.exception("Invalid value for --Depth:", switch[1], lException=e)
        elif switch[0] == "Production":
            try:
                prod = int(switch[1])
            except ValueError as e:
                gLogger.exception("Invalid value for --Production:", switch[1], lException=e)
                prod = 0
        elif switch[0] == "Full":
            full = True
        elif switch[0] == "AllFileTypes":
            sameType = False

    args = Script.getPositionalArgs()

    try:
        level = int(args[-1])
        args.pop()
    except (ValueError, IndexError):
        pass
    if level == 1:
        relation = "NoSister"
    else:
        relation = "NoCousin"

    for lfn in args:
        dmScript.setLFNsFromFile(lfn)
    lfnList = dmScript.getOption("LFNs", [])
    if not lfnList:
        gLogger.error("No LFNs given")
        diracExit(1)

    prodLfns = {}
    if not prod:
        # Get the productions for the input files
        directories = {}
        for lfn in lfnList:
            directories.setdefault(os.path.dirname(lfn), []).append(lfn)
        res = bkClient.getDirectoryMetadata(list(directories))
        if not res["OK"]:
            gLogger.error("Error getting directories metadata", res["Message"])
            diracExit(1)
        for dirName in directories:
            prod = res["Value"]["Successful"].get(dirName, [{}])[0].get("Production")
            if not prod:
                gLogger.error(f"Error: could not get production number for {dirName}")
            else:
                prodLfns.setdefault(prod, []).extend(directories[dirName])
    else:
        prodLfns[prod] = lfnList

    if full:
        resItem = "WithMetadata"
    else:
        resItem = "Successful"
    fullResult = {"OK": True, "Value": {resItem: {}, relation: set()}}
    resValue = fullResult["Value"]

    for prod, lfnList in prodLfns.items():  # can be an iterator
        if sameType:
            res = bkClient.getFileMetadata(lfnList)
            if not res["OK"]:
                gLogger.error("Error getting file metadata", res["Message"])
                diracExit(1)
            lfnTypes = {}
            for lfn in res["Value"]["Successful"]:
                metadata = res["Value"]["Successful"][lfn]
                lfnTypes[lfn] = metadata["FileType"]
        else:
            lfnTypes = dict.fromkeys(lfnList, None)

        # First get ancestors
        result = bkClient.getFileAncestors(list(lfnTypes), level, replica=False)
        if not result["OK"]:
            gLogger.error("Error getting ancestors:", res["Message"])
            diracExit(1)

        ancestors = {}
        # More than one file in the input list may have the same ancestor(s), check if they are sisters/cousins
        for lfn, ancList in result["Value"]["Successful"].items():  # can be an iterator
            sameAncestors = {anc["FileName"] for anc in ancList} & set(ancestors)
            skip = False
            if sameAncestors:
                # Some of the input files share the same ancestor, are they sisters/cousins?
                for sameAnc in sameAncestors:
                    for desc in ancestors[sameAnc]:
                        # Check if this file has the same type as lfn: if so, they are sisters/cousins, then keep only one
                        if lfnTypes[lfn] == lfnTypes[desc]:
                            skip = True
            if not skip:
                for anc in ancList:
                    ancestors.setdefault(anc["FileName"], []).append(lfn)
        # print ancestors

        res = bkClient.getFileDescendents(list(ancestors), depth=999999, production=prod, checkreplica=checkreplica)

        fullResult["OK"] = res["OK"]
        if res["OK"]:
            for anc, sisters in res["Value"]["WithMetadata"].items():  # can be an iterator
                lfns = ancestors[anc]
                found = False
                for sister in sisters:
                    metadata = sisters[sister]
                    if sister not in lfns and (not sameType or metadata["FileType"] in (lfnTypes[lfn] for lfn in lfns)):
                        allLfns = ", ".join(lfns)
                        if full:
                            resValue[resItem].setdefault(allLfns, {}).update(metadata)
                        else:
                            resValue[resItem].setdefault(allLfns, {}).update(
                                {sister: f"Replica-{metadata['GotReplica']}"}
                            )
                        found = True
                for lfn in lfns:
                    if not found:
                        resValue[relation].add(lfn)
                    if lfn in lfnList:
                        lfnList.remove(lfn)
            for lfn in lfnList:
                resValue[relation].add(lfn)
            for lfn in set(resValue[resItem]).intersection(resValue[relation]):
                resValue[relation].remove(lfn)
        else:
            break

    diracExit(printDMResult(fullResult, empty="None", script="dirac-bookkeeping-get-file-sisters"))


def _intWithQuotes(val, quote="'"):
    """Print numbers with a character separating each thousand.

    :param int,long val: integer value to printout
    :param character quote: character to use as a separator

    :return: string to print out
    """
    chunks = []
    if not val:
        return "None"
    while val:
        if val >= 1000:
            chunks.append("%03d" % (val % 1000))
        else:
            chunks.append("%d" % (val % 1000))
        val = int(val / 1000)
    chunks.reverse()
    return quote.join(chunks)


def _scaleValue(val, units, factor=1000):
    """Scale a value by thousands, return value and unit.

    :param float val: value to scale
    :param iterable units: list of unit names (increasing order)

    :return: tuple (scaledValue, unitName)
    """
    if val:
        for unit in units:
            if val < factor:
                break
            val /= factor
    else:
        val = 0
        unit = ""
    return val, unit


def _scaleLumi(lumi):
    """Return lumi in the appropriate unit."""
    return _scaleValue(lumi, ("/microBarn", "/nb", "/pb", "/fb", "/ab"))


def scaleSize(size):
    """Return size in appropriate unit."""
    return _scaleValue(size, ("Bytes", "kB", "MB", "GB", "TB", "PB"))


def scaleiSize(size):
    """Return size in appropriate unit."""
    return _scaleValue(size, ("Bytes", "kiB", "MiB", "GiB", "TiB", "PiB"), factor=1024)


def _getCollidingBunches(fills):
    """Get the number of colliding bunches for all fills and average.

    :param iterable fills: list of fill numbers

    :return: dictionary {fill:nbCollisingBunches}
    """
    import json
    from urllib.request import urlopen
    from urllib.error import HTTPError

    result = {}
    for fill in fills:
        try:
            runDbUrl = "https://lbrundb.cern.ch/api/fill/%d/" % fill
            fillInfo = json.load(urlopen(runDbUrl))
            result[fill] = int(fillInfo["nCollidingBunches"])
        except (KeyError, ValueError) as e:
            gLogger.exception("Exception getting info for fill", str(fill), lException=e)
        except HTTPError:
            pass
    return result


def executeGetStats(dmScript):
    """Extract statistics for a BK query or a set of LFNs.

    :param dmScript: instance containing options (LFNs or BKQuery mandatory)
    :type dmScript: DMScript class
    """
    triggerRate = False
    listRuns = False
    listFills = False
    paramNames = []
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "TriggerRate":
            triggerRate = True
        elif switch[0] == "ListRuns":
            listRuns = True
        elif switch[0] == "ListFills":
            listFills = True

    if listRuns:
        listFills = False
    lfns = dmScript.getOption("LFNs")
    bkQuery = dmScript.getBKQuery()
    if not bkQuery and not lfns:
        gLogger.notice("No BK query given...")
        diracExit(1)

    if bkQuery:
        prodList = bkQuery.getQueryDict().get("Production", [None])
        if not isinstance(prodList, list):
            prodList = [prodList]
        bkQuery.setOption("ProductionID", None)
        fileType = bkQuery.getFileTypeList()
        processingPass = bkQuery.getProcessingPass().replace("...", "*")
        if "*" in processingPass:
            depth = None if processingPass.endswith("/*") else 1
            progressBar = ProgressBar(1, title=f"Getting list of processing passes (depth={depth})")
            processingPasses = getProcessingPasses(bkQuery, depth=depth)
            progressBar.endLoop(message=f"obtained {len(processingPasses)} processing passes")
        else:
            processingPasses = [processingPass]
    else:
        prodList = [None]
        fileType = None
    if triggerRate and fileType != ["RAW"] and not [lfn for lfn in lfns if ".raw" in lfn]:
        gLogger.notice("TriggerRate option ignored as not looking for RAW files")
        triggerRate = False

    transClient = TransformationClient()

    for prod in prodList:
        # Get bk query dict if any, no change to default values
        queryDict = bkQuery.getQueryDict() if bkQuery else {}
        if prod:
            res = transClient.getTransformation(prod, extraParams=False)
            if not res["OK"]:
                continue
            prodName = res["Value"]["TransformationName"]
            bkQuery.setOption("Production", prod)
            gLogger.notice("For production %d, %s (query %s)" % (prod, prodName, bkQuery))
        elif lfns:
            gLogger.notice(f"For {len(lfns)} LFNs:")
        elif bkQuery:
            queryDict = bkQuery.getQueryDict()
            if len(processingPasses) > 1:
                queryDict.pop("ProcessingPass", None)
                gLogger.notice("For BK query:", queryDict)
                gLogger.notice(f"   and {len(processingPasses)} processing passes :", ", ".join(processingPasses))
            else:
                gLogger.notice("For BK query:", bkQuery)

        # Get information from BK
        if not triggerRate and not lfns:
            #
            # Except if we need to get the list of files, use getFilesSummary
            #
            fileTypes = queryDict.pop("FileType", None)
            if not isinstance(fileTypes, list):
                fileTypes = [fileTypes]
            records = []
            nDatasets = 0
            eventTypes = queryDict.get("EventType")
            nd = len(processingPasses) * len(fileTypes)
            progressBar = ProgressBar(
                nd, title="Getting info from filesSummary" + (" for %d datasets..." % nd if nd > 1 else "..."), step=1
            )
            for processingPass in processingPasses:
                # Loop on all processing passes if needed
                if processingPass:
                    queryDict["ProcessingPass"] = processingPass
                for fileType in fileTypes:
                    progressBar.loop()
                    if fileType:
                        queryDict["FileType"] = fileType
                    retry = 0
                    while retry < 5:
                        retry += 1
                        res = bkClient.getFilesSummary(queryDict)
                        if res["OK"]:
                            break
                    if not res["OK"]:
                        gLogger.error("Error getting statistics from BK", res["Message"])
                        diracExit(1)
                    paramNames = res["Value"]["ParameterNames"]
                    record = len(paramNames) * [0]
                    for paramValues in res["Value"]["Records"]:
                        record = [(rec + val) if val else rec for rec, val in zip(record, paramValues)]
                    # print fileType, record
                    recDict = dict(zip(paramNames, record))
                    nevts = recDict.get("NumberOfEvents", 0)
                    if nevts:
                        nDatasets += len(eventTypes) if isinstance(eventTypes, list) else 1
                    size = recDict.get("FileSize", 0)
                    lumi = recDict.get("Luminosity", 0)
                    record.append(nevts / float(lumi) if lumi else 0.0)
                    record.append(size / float(lumi) if lumi else 0.0)
                    if not records:
                        records = len(record) * [0]
                    records = [rec1 + rec2 for rec1, rec2 in zip(record, records)]
                    # print fileType, records, 'Total'
            progressBar.endLoop()
            paramNames += ["EvtsPerLumi", "SizePerLumi"]
            nDatasets = max(1, nDatasets)
            # End of getFilesSummary section
        else:
            #
            # Here it is because we really need the list of files
            #
            paramNames = ["NbofFiles", "NumberOfEvents", "FileSize", "Luminosity", "EvtsPerLumi", "SizePerLumi"]
            nbFiles = 0
            nbEvents = 0
            fileSize = 0
            lumi = 0
            datasets = set()
            runList = {}
            if not lfns:
                nDatasets = len(processingPasses)
                title = "Getting info from files" + (" for %d datasets..." % nDatasets if nDatasets > 1 else "...")
                progressBar = ProgressBar(nDatasets, title=title, step=1)
                for processingPass in processingPasses:
                    if processingPass:
                        queryDict["ProcessingPass"] = processingPass
                    progressBar.loop()
                    fileTypes = queryDict.get("FileType")
                    retry = 0
                    while retry < 5:
                        retry += 1
                        res = bkClient.getFilesWithMetadata(queryDict)
                        if res["OK"]:
                            break
                    if not res["OK"]:
                        progressBar.endLoop(message="Fatal error")
                        gLogger.fatal("Error getting files with metadata", res["Message"])
                        diracExit(1)
                    if "ParameterNames" not in res["Value"]:
                        progressBar.endLoop(message="Fatal error")
                        gLogger.fatal(
                            "ParameterNames not present:",
                            ",".join(sorted(res["Value"])) if isinstance(res["Value"], dict) else str(res["Value"]),
                        )
                        diracExit(1)
                    parameterNames = res["Value"]["ParameterNames"]
                    info = res["Value"]["Records"]
                    for item in info:
                        metadata = dict(zip(parameterNames, item))
                        try:
                            if metadata["EventStat"]:
                                datasets.add(
                                    (
                                        metadata.get("EventType", metadata["FileName"].split("/")[5]),
                                        metadata.get("FileType", metadata.get("Name")),
                                    )
                                )
                            fileSize += metadata["FileSize"]
                            lumi += metadata["Luminosity"]
                            run = metadata["RunNumber"]
                            runList.setdefault(run, [0.0, 0.0, 0.0])
                            runList[run][0] += metadata["Luminosity"]
                            if metadata["EventStat"]:
                                nbEvents += metadata["EventStat"]
                                runList[run][1] += metadata["EventStat"]
                            runList[run][2] += metadata["FileSize"]
                            nbFiles += 1
                        except (KeyError, ValueError) as e:
                            gLogger.exception(f"Exception for {str(list(metadata))}", lException=e)
                progressBar.endLoop()
            if lfns:
                gLogger.notice("Getting info from files...")
                lfnChunks = breakListIntoChunks(lfns, 1000)
                progressBar = ProgressBar(len(lfns), title="Get metadata from BK", chunk=1000)
                for lfnChunk in lfnChunks:
                    progressBar.loop()
                    res = bkClient.getFileMetadata(lfnChunk)
                    if res["OK"]:
                        for lfn, metadata in res["Value"]["Successful"].items():  # can be an iterator
                            try:
                                if metadata["EventStat"]:
                                    datasets.add((metadata["EventType"], metadata["FileType"]))
                                fileSize += metadata["FileSize"]
                                lumi += metadata["Luminosity"]
                                run = metadata["RunNumber"]
                                runList.setdefault(run, [0, 0, 0])
                                runList[run][0] += metadata["Luminosity"]
                                if metadata["EventStat"]:
                                    nbEvents += metadata["EventStat"]
                                    runList[run][1] += metadata["EventStat"]
                                runList[run][2] += metadata["FileSize"]
                                nbFiles += 1
                            except (KeyError, ValueError) as e:
                                gLogger.exception(f"Exception for {lfn}", str(list(metadata)), lException=e)
                    else:
                        gLogger.error("Error getting files metadata:", res["Message"])
                        continue
                progressBar.endLoop()
            records = [
                nbFiles,
                nbEvents,
                fileSize,
                lumi,
                nbEvents / float(lumi) if lumi else 0.0,
                fileSize / float(lumi) if lumi else 0.0,
            ]
            nDatasets = max(1, len(datasets))
            # End of get info from files section
        #
        # Now printout the results
        #
        tab = 17
        nfiles = nevts = lumi = 0
        for name, value in zip(paramNames, records):
            if name == "NbofFiles":
                nfiles = value
                gLogger.notice("Nb of Files".ljust(tab) + ": ", f"{_intWithQuotes(value)}")
            elif name == "NumberOfEvents":
                nevts = value
                gLogger.notice("Nb of Events".ljust(tab) + ": ", f"{_intWithQuotes(value)}")
            elif name == "FileSize":
                size_bytes = value
                sizePerEvt = f"({size_bytes / nevts / 1000.0:.1f} kB per evt)" if nevts and nDatasets == 1 else ""
                size, sizeUnit = scaleSize(size_bytes)
                gLogger.notice("Total size".ljust(tab) + ": ", f"{size:.3f} {sizeUnit} {sizePerEvt}")
                isize, isizeUnit = scaleiSize(size_bytes)
                isizePerEvt = f"({size_bytes / nevts / 1024.0:.1f} kiB per evt)" if nevts and nDatasets == 1 else ""
                gLogger.notice("Total size (1024)".ljust(tab) + ": ", f"{isize:.3f} {isizeUnit} {isizePerEvt}")
            elif name == "Luminosity":
                lumi = value / nDatasets
                lumi, lumiUnit = _scaleLumi(lumi)
                lumiString = "Luminosity" if nDatasets == 1 else "Avg luminosity"
                gLogger.notice(f"{lumiString.ljust(tab)}:  {lumi:.3f} {lumiUnit}")
            elif name == "SizePerLumi":
                # value *= nDatasets
                gLogger.notice("Size per /pb".ljust(tab) + ": ", f"{value * 1e-3:.1f} GB")
        if lumi:
            filesPerLumi = nfiles / lumi
            gLogger.notice(f"Files per {lumiUnit}".ljust(tab) + ": ", f"{filesPerLumi:.1f}")

        if triggerRate:
            # Get information from the runs, but first get those that are Finished
            res = bkClient.getRunStatus(list(runList))
            if not res["OK"]:
                gLogger.error("Error getting run status", res["Message"])
                runs = []
            else:
                success = res["Value"]["Successful"]
                runs = [run for run in success if success[run].get("Finished") == "Y"]
            notFinished = set(runList) - set(runs)
            if notFinished:
                gLogger.notice(
                    f"{len(notFinished)} runs not Finished (ignored), "
                    f"{len(runs) if runs else 'no'} runs Finished (used for trigger rate)"
                )
                gLogger.notice("These runs are not Finished: ", ",".join(str(run) for run in sorted(notFinished)))
            if runs:
                nevts = 0
                size = 0
                fullDuration = 0.0
                totalLumi = 0.0
                fills = {}
                fillDuration = {}
                for run in sorted(runs):
                    res = bkClient.getRunInformations(run)
                    if not res["OK"]:
                        gLogger.error("Error from BK getting run information", res["Message"])
                    else:
                        nevts += runList[run][1]
                        size += runList[run][2]
                        info = res["Value"]
                        fill = info["FillNumber"]
                        fills.setdefault(fill, []).append(str(run))
                        runDuration = (info["RunEnd"] - info["RunStart"]).total_seconds() / 3600.0
                        fillDuration[fill] = fillDuration.setdefault(fill, 0) + runDuration
                        fullDuration += runDuration
                        lumi = info["TotalLuminosity"]
                        if abs(lumi - runList[run][0] / nDatasets) > 1:
                            gLogger.notice(
                                "Run and files luminosity mismatch (ignored): ",
                                f"run {run}, runLumi {lumi}, filesLumi {runList[run][0] / nDatasets}",
                            )
                        else:
                            totalLumi += lumi
                if fullDuration:
                    triggerRate = nevts / fullDuration / 3600
                    rate = f"{triggerRate:.1f} events/second"
                else:
                    triggerRate = 0.0
                    rate = "Run duration not available"
                totalLumi, lumiUnit = _scaleLumi(totalLumi)
                gLogger.notice(f"{'Total Luminosity'.ljust(tab)}: {totalLumi:.3f} {lumiUnit}")
                gLogger.notice(f"{'Run duration'.ljust(tab)}: {fullDuration:.2f} hours ({len(runs)} runs)")
                gLogger.notice(f"{'Trigger rate'.ljust(tab)}: {rate}")
                rate = (
                    f"{size / 1000000.0 / fullDuration / 3600.0:.1f} MB/second"
                    if fullDuration
                    else "Run duration not available"
                )
                gLogger.notice(f"{'Throughput'.ljust(tab)}: {rate}")
                collBunches = 0.0
                result = {}
                # FIXME: if/when the online run DB is accessible to get the number of bunches this should be re-activated
                # result = _getCollidingBunches(fillDuration)
                # for fill in fillDuration:
                #   if fill not in result:
                #     gLogger.notice("Error: no number of colliding bunches for fill %d" % fill)
                #   else:
                #     collBunches += result[fill] * fillDuration[fill]
                if fullDuration and collBunches:
                    collBunches /= fullDuration
                    gLogger.notice(f"{'Colliding bunches'.ljust(tab)}: {collBunches:.1f} on average")
                    gLogger.notice(f"{'Trigger per bunch'.ljust(tab)}: {triggerRate / collBunches:.2f} events/s/bunch")
                if listFills:
                    gLogger.notice(
                        "List of fills: ",
                        ",".join(
                            "%d (%d runs, %.1f hours)" % (fill, len(fills[fill]), fillDuration[fill])
                            for fill in sorted(fills)
                        ),
                    )
                if listRuns:
                    for fill in sorted(fills):
                        gLogger.notice(
                            "Fill %d (%s, %.1f hours):"
                            % (
                                fill,
                                "%4d bunches" % result[fill] if fill in result else "Unknown bunches",
                                fillDuration[fill],
                            ),
                            ",".join(fills[fill]),
                        )
        gLogger.notice("")


def executeRunInfo(item):
    """Get some run information for a range of runs, either print out by item
    value or by run range It can also define run ranges according to a run number
    gap or a time gap.

    :param item: name of a run item
    :type item: string
    """
    runsDict = {}
    if item not in ("Tck", "DataTakingDescription", "ProcessingPass", "Ranges"):
        gLogger.fatal("Incorrect run information item")
        return 0
    force = None
    byRange = False
    activity = None
    dqFlag = None
    runGap = 100
    timeGap = 365
    eventType = 90000000
    for switch in Script.getUnprocessedSwitches():
        if switch[0] in ("Force", "Fast"):
            # Set True only if not changed by another option
            if force is None:
                force = True
        elif switch[0] == "ByRange":
            byRange = True
        elif switch[0] == "Activity":
            activity = switch[1].capitalize()
        elif switch[0] == "EventType":
            eventType = int(switch[1])
        elif switch[0] == "DQFlag":
            dqFlag = switch[1].split(",")
        elif switch[0] == "RunGap":
            runGap = int(switch[1])
        elif switch[0] == "TimeGap":
            # Must get run information to apply it
            timeGap = int(switch[1])
            runGap = 100000000
            force = False
        elif switch[0] == "Runs":
            # Add a fake run number to force parseRuns to return a list
            try:
                runRange = switch[1]
                runsDict = parseRuns({}, "0," + runRange)
            except BadRunRange:
                gLogger.fatal("Bad run range")
                diracExit(1)

    getRanges = item == "Ranges"
    byRange = byRange or getRanges
    if not getRanges:
        # Avoid creating gaps if explicit run ranges are not requested
        runGap = 100000000
        timeGap = 365
    if activity:
        progressBar = ProgressBar(1, title=f"Getting run list for activity {activity}", step=20)
        runsList = BKQuery(f"/LHCb/{activity}//Real Data/{eventType}/RAW").getBKRuns()
        progressBar.endLoop(message=f"Obtained {len(runsList)} runs")
        if runsDict:
            runsList = sorted(set(runsList) & set(runsDict["RunNumber"]))
            gLogger.notice(f"Only {len(runsList)} runs in range {runRange}")
    elif not runsDict:
        Script.showHelp()
        diracExit(1)
    else:
        runsList = runsDict["RunNumber"]
        # Remove the fake run number
        runsList.remove(0)
        if len(runRange) < 30:
            progressBar = ProgressBar(1, title=f"Getting runs for run range {runRange} ", step=20)
        else:
            progressBar = ProgressBar(1, title=f"Getting runs for {len(runsList)} runs ", step=20)
        res = bkClient.getRunStatus(runsList)
        progressBar.endLoop()
        if not res["OK"]:
            gLogger.fatal("Error getting run list", res["Message"])
            diracExit(1)
        runsList = list(res["Value"]["Successful"])

    if dqFlag:
        nruns = len(runsList)
        chunkSize = 10
        progressBar = ProgressBar(len(runsList), title=f"Getting DQFlag for {len(runsList)} runs ", chunk=chunkSize)
        for runChunk in breakListIntoChunks(runsList, chunkSize):
            progressBar.loop()
            res = bkClient.getRunFilesDataQuality(runChunk)
            if not res["OK"]:
                gLogger.fatal("Error getting DQFlag", res["Message"])
                diracExit(1)
            dq = {r: dq for r, dq, s in res["Value"] if s == 90000000}
            for run in runChunk:
                if dq.get(run) not in dqFlag:
                    runsList.remove(run)
        progressBar.endLoop()
        if len(runsList) != nruns:
            gLogger.notice(f"Only {len(runsList)} runs have DQ flag in {','.join(dqFlag)}")

    runTime = {}
    if getRanges and force:
        counted = "runs"
        runDict = dict.fromkeys(runsList, 1)
    else:
        counted = "files"
        runDict = {}
        progressBar = ProgressBar(len(runsList), title=f"Getting {item} for {len(runsList)} runs ", step=20)
        for run in sorted(runsList):
            progressBar.loop()
            res = bkClient.getRunInformations(run)
            if res["OK"]:
                condition = res["Value"]["DataTakingDescription"]
                if "Excl" in condition and getRanges:
                    continue
                streams = res["Value"].get("Stream", [])
                runTime[run] = (res["Value"].get("RunStart"), res["Value"].get("RunEnd"), condition)
                if getRanges:
                    files = res["Value"].get("Number of file", [])
                    itemValue = dict(zip(streams, files)).get(90000000, 0)
                else:
                    itemValue = res["Value"].get(item)
                if (90000000 in streams or force) and itemValue:
                    runDict[run] = itemValue
        progressBar.endLoop()
    if getRanges:
        itemList = [1]
    else:
        itemList = []
        for run in sorted(runDict):
            if runDict[run] not in itemList:
                itemList.append(runDict[run])
    itemDict = {}
    rangesDict = {}
    runList = sorted(runDict)
    for itemValue in itemList:
        firstRun = None
        lastRun = None
        runStart = None
        lastRunEnd = None
        lastRunDesc = None
        lastRunValue = None
        if getRanges:
            count = 0
        # Add a fake run (None) in order to print out the last range
        if itemValue == itemList[-1]:
            runList.append(None)
        for run in runList:
            runValue = 1 if getRanges and run else runDict.get(run)
            if runValue == lastRunValue and runValue != itemValue:
                # We are outside a range that is interesting for that itemValue, skip
                continue
            runTimeAndDesc = runTime.get(run, (None, None, None))
            runStart = runTimeAndDesc[0]
            runDesc = runTimeAndDesc[2]
            # Determine if there is a need to change the run range
            # Can be a change of conditions, a runGap in time or a runGap in #of runs
            gap = False
            if lastRunDesc:
                gap = runDesc != lastRunDesc
            if not gap and runStart and lastRunEnd:
                runDiff = runStart - lastRunEnd
                gap = runDiff > datetime.timedelta(days=timeGap)
            if not gap and lastRun and run:
                gap = abs(lastRun - run) > runGap

            if runValue == itemValue and firstRun is None:
                # First run encountered
                firstRun = run
                if getRanges:
                    # Initialize count of files
                    count = 0
            elif (runValue != itemValue or gap) and firstRun is not None:
                # We are now in a new range, print out the previous range
                if lastRun != firstRun:
                    # This should be "disable=bad-string-format-type" but this
                    # isn't support in Python 2 versions of pylint
                    rangeStr = f"{firstRun}:{lastRun}"
                else:
                    rangeStr = f"{firstRun}"
                if lastRunDesc:
                    rangeStr += f" ({lastRunDesc})"
                if getRanges:
                    rangesDict[rangeStr] = f"{count} {counted}"
                    # Initialize count of files
                    count = 0
                else:
                    rangesDict[rangeStr] = itemValue
                itemDict.setdefault(itemValue, []).append(rangeStr)
                # If still same value, start a new range
                firstRun = run if runValue == itemValue else None
            if getRanges and run:
                count += runDict[run]
            # Update parameters with this run's information
            lastRun = run
            lastRunDesc = runDesc
            lastRunEnd = runTimeAndDesc[1]
            lastRunValue = runValue

    if byRange:
        if item == "Ranges":
            gLogger.notice(f"Total number of runs: {len(runDict)}")
        for rangeStr in sorted(rangesDict):
            gLogger.notice(f"{rangeStr} : {rangesDict[rangeStr]}")
    else:
        for itemValue in sorted(itemDict):
            gLogger.notice(f"{itemValue} :", ", ".join(itemDict[itemValue]))


def _jobFromLfn(lfn):
    """Extract job unique name form LFN."""
    return os.path.basename(lfn).split(".")[0]


def _getJobsEISFromAncestors(lfnList):
    """Get EventInputStat of all jobs in a recursive way The EventInputStat of
    all jobs is set in the global dictionary jobEventInputStat.

    :param lfnList: list of LFNs
    """
    # Get one output file per unknown job
    jobDict = {}
    for lfn in lfnList:
        job = _jobFromLfn(lfn)
        if job not in jobEventInputStat:
            jobDict.setdefault(job, lfn)
    lfnList = list(jobDict.values())
    # Get ancestors of these files
    gLogger.verbose(f"\nGet EIS for {len(jobDict)} jobs like {str(list(jobDict.items())[0])}")
    gLogger.verbose(f"\t{len(set(lfnList))} unique files")
    ancToCheck = set()
    ancestors = {}
    topFiles = 0
    for lfnChunk in breakListIntoChunks(lfnList, 100):
        res = bkClient.getFileAncestors(lfnChunk, depth=1, replica=False)
        if not res["OK"]:
            return res
        ancWithMetadata = res["Value"]["WithMetadata"]
        for lfn, ancDict in ancWithMetadata.items():  # can be an iterator
            job = _jobFromLfn(lfn)
            for anc, meta in ancDict.items():  # can be an iterator
                ancJob = _jobFromLfn(anc)
                ancestors.setdefault(job, []).append(ancJob)
                if ancJob in jobEventInputStat:
                    # This ancestor job is  already known
                    jobEventInputStat[job] = jobEventInputStat.setdefault(job, 0) + jobEventInputStat[ancJob]
                else:
                    top = meta["FileType"] in ("SIM", "RAW")
                    if top:
                        # We are at the top, the EventInputStat is the sum of EventStat of input files
                        topFiles += 1
                        jobEventInputStat[job] = jobEventInputStat.setdefault(job, 0) + meta["EventStat"]
                    else:
                        ancToCheck.add(anc)

    if topFiles:
        gLogger.info("Found %d top files" % topFiles)
    if ancToCheck:
        # Let's go one step further
        res = _getJobsEISFromAncestors(list(ancToCheck))
        if not res["OK"]:
            return res
        # Update the table for jobs still unknown
        for job, ancJobs in ancestors.items():  # can be an iterator
            if job not in jobEventInputStat:
                try:
                    # This ancestor job is  already known
                    jobEventInputStat[job] = sum(jobEventInputStat[ancJob] for ancJob in ancJobs)
                except KeyError as e:
                    gLogger.exception("ERROR: ancestor job not found in table", lException=e)

    return S_OK()


def _getEventInputStat(lfns):
    """Get EventInputStat of all ancestor jobs of a list of files."""
    # Sort events by job
    jobLfns = {}
    for lfn in lfns:
        jobName = _jobFromLfn(lfn)
        if jobName not in jobEventInputStat:
            jobLfns.setdefault(jobName, []).append(lfn)
    # These are the final LFNs, but we must get to the top ancestors to get it right
    # Let's do it job by job
    if jobLfns:
        progressBar = ProgressBar(len(jobLfns), title=f"Getting ancestors for {len(jobLfns)} jobs", chunk=1)
        for lfns in jobLfns.values():
            progressBar.loop()
            res = _getJobsEISFromAncestors(lfns)
            if not res["OK"]:
                return res
        progressBar.endLoop()
    return S_OK()


def _checkEventInputStat(lfn):
    """Check if EventInputStat is correct."""
    progressBar = ProgressBar(1, title="Checking if BK EventInputStat is reliable...", chunk=1)
    # Get EventInputStat from ancestors
    res = _getJobsEISFromAncestors([lfn])
    if not res["OK"]:
        progressBar.endLoop(message="*** Error in _getJobsEISFromAncestors: " + res["Message"] + "***")
        return res
    # Get EventInputStat for one file and keep the information in case it is useful
    topInputStat = jobEventInputStat[_jobFromLfn(lfn)]
    # Get EventInputStat from the parent job
    res = bkClient.getJobInfo(lfn)
    if not res["OK"]:
        return res
    eventInputStat = res["Value"][0][2]
    if eventInputStat != topInputStat:
        gLogger.info("EventInputStat mismatch: %d in job, %d from ancestors" % (eventInputStat, topInputStat))
        result = False
    else:
        result = True
    progressBar.endLoop(message="OK" if result else "not OK, get info from ancestors")
    return S_OK(result)


def executeRejectionStats(dmScript):
    """Get rejection rate for stripping streams.

    :param dmScript: instance containing options (LFNs or BKQuery mandatory)
    :type dmScript: DMScript class
    """
    byStream = False
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "ByStream":
            byStream = True

    lfns = dmScript.getOption("LFNs", [])
    bkQuery = dmScript.getBKQuery()
    if not bkQuery and not lfns:
        gLogger.notice("No BK query given...")
        diracExit(1)

    fileTypes = [""]
    if bkQuery:
        # If looking for invisible files, get all of them
        rFlag = bkQuery.isVisible()
        if rFlag == "No":
            rFlag = "All"
        bkQuery.setOption("ReplicaFlag", rFlag)
        fileTypes = bkQuery.getFileTypeList()
        if not fileTypes:
            fileTypes = bkQuery.getBKFileTypes()
        gLogger.notice("Using BK query", str(bkQuery))
        if byStream:
            streams = sorted(fileTypes)
        else:
            streams = ["all"]

    else:
        streams = ["all"]

    chunkSize = 100
    eventStatByStream = {}
    eventInputStatByStream = {}
    uniqueJobs = {}
    lfnsByStream = {}
    jobs = {}
    for stream in streams:
        if bkQuery:
            if byStream:
                bkQuery.setFileType(stream)
            lfns = bkQuery.getLFNs(printSEUsage=False, printOutput=False)

        if not lfns:
            gLogger.notice("No files found")
            if not byStream:
                diracExit(0)
            continue

        # Make a quick check that EventInputStat is correct
        res = _checkEventInputStat(sorted(lfns)[0])
        if not res["OK"]:
            gLogger.fatal("Error checking EventInputStat", res["Message"])
            diracExit(1)
        evtInputStatOK = res["Value"]

        lfnsByStream[stream] = lfns
        progressBar = ProgressBar(
            len(lfns),
            title="Getting metadata for %d files %s" % (len(lfns), ("(stream %s)" % stream) if byStream else ""),
            chunk=chunkSize,
        )
        eventStat = 0
        for lfnChunk in breakListIntoChunks(lfns, chunkSize):
            progressBar.loop()
            for lfn in lfnChunk:
                job = _jobFromLfn(lfn)
                if job not in jobEventInputStat:
                    uniqueJobs.setdefault(job, lfn)
            res = bkClient.getFileMetadata(lfnChunk)
            if not res["OK"]:
                gLogger.fatal("Error getting files metadata", res["Message"])
                diracExit(1)
            eventStat += sum(meta["EventStat"] for meta in res["Value"]["Successful"].values() if meta["EventStat"])
        progressBar.endLoop()
        eventStatByStream[stream] = eventStat

        jobLfns = set(uniqueJobs.values()) & set(lfns)
        if jobLfns:
            if not evtInputStatOK:
                gLogger.notice("EventInputStat cannot be trusted, it will take a bit more time...")
                res = _getEventInputStat(jobLfns)
                if not res["OK"]:
                    gLogger.fatal("Error getting ancestors EventStat", res["Message"])
                    diracExit(1)
            else:
                progressBar = ProgressBar(len(lfns), title=f"Getting metadata for {len(jobLfns)} jobs", chunk=chunkSize)
                for lfnChunk in breakListIntoChunks(jobLfns, chunkSize):
                    progressBar.loop()
                    res = bkClient.bulkJobInfo(lfnChunk)
                    if not res["OK"]:
                        gLogger.fatal("Error getting job information", res["Message"])
                        diracExit(1)
                    success = res["Value"]["Successful"]
                    for lfn in success:
                        if isinstance(success[lfn], list) and len(success[lfn]) == 1:
                            success[lfn] = success[lfn][0]
                        jobName = _jobFromLfn(lfn)
                        jobEventInputStat[jobName] = success[lfn]["EventInputStat"]
                progressBar.endLoop()

        eventInputStatByStream[stream] = sum(jobEventInputStat[_jobFromLfn(lfn)] for lfn in lfns)
        jobs[stream] = {_jobFromLfn(lfn) for lfn in lfns}
    # End of loop on streams

    tab = "\t" if byStream else ""
    eventInputStat = 0
    multiStream = len(fileTypes) > 1 and not byStream and len(jobs["all"]) == len(lfnsByStream["all"])
    if multiStream:
        gLogger.notice(f"*** WARNING *** {len(fileTypes)} streams with as many jobs as files, rescaled input stat")
        rescale = len(fileTypes)
    else:
        rescale = 1
    for stream in streams:
        eventStat = eventStatByStream[stream]
        prInput = eventInputStat != eventInputStatByStream[stream]
        if prInput:
            eventInputStat = eventInputStatByStream[stream] / float(rescale)
        if len(fileTypes) > 1:
            gLogger.notice(f"For {stream} stream{'' if byStream else 's'}:")
        gLogger.notice(tab + "Event stat: %d on %d files" % (eventStat, len(lfnsByStream[stream])))
        if prInput:
            gLogger.notice(
                tab + "EventInputStat: %d" % eventInputStat,
                f"from {len(jobs[stream])} jobs" if jobs[stream] != lfnsByStream[stream] else "",
            )
        gLogger.notice(tab + f"Retention: {100.0 * eventStat / eventInputStat:.2f} %")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
    This module implements the default behavior for the FTS3Agent for TPC and source SE selection
"""
import random
import itertools
import re
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.DataManagementSystem.private.FTS3Plugins.DefaultFTS3Plugin import DefaultFTS3Plugin
from DIRAC.Resources.Storage.StorageElement import StorageElement

sLog = gLogger.getSubLogger(__name__)


class LHCbFTS3Plugin(DefaultFTS3Plugin):
    @staticmethod
    def _isRalAntaresEchoStaging(sourceSEName=None, destSEName=None):
        """Check if the transfer goes from Antares to Echo at RAL"""
        try:
            srcSE = StorageElement(sourceSEName)
            srcBaseSEName = srcSE.options.get("BaseSE")
            dstSE = StorageElement(destSEName)
            dstBaseSEName = dstSE.options.get("BaseSE")

            if srcBaseSEName in ("RAL-ANTARES-DATACHALLENGE", "RAL-ANTARES"):
                if dstBaseSEName in ("RAL-ECHO", "RAL-ECHO-DATACHALLENGE"):
                    return True

        except Exception:
            pass

        return False

    def selectTPCProtocols(self, ftsJob=None, sourceSEName=None, destSEName=None, **kwargs):
        """Specialised TPC selection"""

        if not sourceSEName:
            sourceSEName = ftsJob.sourceSE
        if not destSEName:
            destSEName = ftsJob.targetSE

        # # if we stage from Antares to Echo, return root as TPC
        # if self._isRalAntaresEchoStaging(sourceSEName=sourceSEName, destSEName=destSEName):
        #     return ["root"]

        return super().selectTPCProtocols(ftsJob=ftsJob, sourceSEName=sourceSEName, destSEName=destSEName, **kwargs)

    def selectSourceSE(self, ftsFile, replicaDict, allowedSources):
        """
        This is basically a copy/paste of the parent method, with the exception
        of prefering local staging.
        """

        allowedSourcesSet = set(allowedSources) if allowedSources else set()

        # If we have a restriction, apply it, otherwise take all the replicas
        allowedReplicaSource = (set(replicaDict) & allowedSourcesSet) if allowedSourcesSet else replicaDict
        if not allowedReplicaSource:
            sLog.warn(
                "Incompatible restriction, trying to find more replicas",
                f"{ftsFile.lfn=}, {replicaDict=}, {allowedSources=}",
            )

            # Query again the replicas in order to prefer disks
            # as this is not the case when specifying preferd sources
            allowedReplicaSource = returnSingleResult(
                DataManager().getActiveReplicas(ftsFile.lfn, getUrl=False, preferDisk=True)
            ).get("Value")
            if not allowedReplicaSource:
                sLog.warn(
                    "Could not retrieve replicas again, taking all replicas",
                    f"{ftsFile.lfn=} {replicaDict=}, {allowedSources=}",
                )
                allowedReplicaSource = replicaDict

        # If we have a replica at the same site as the destination
        # use that one
        # This is mostly done in order to favor local staging
        #
        # We go with the naive assumption that the site name
        # is always the first part of the SE name, separated
        # by either a - or a _ (like `_MC-DST`)
        # (I know there are "proper tools" for checking if a SE is on the same site
        # but since we are in the sheltered LHCb only environment, I can do that
        # sort of optimization)
        targetSite = re.split("-|_", ftsFile.targetSE)[0]
        sameSiteSE = [srcSE for srcSE in allowedReplicaSource if targetSite in srcSE]
        if sameSiteSE:
            allowedReplicaSource = sameSiteSE

        randSource = random.choice(list(allowedReplicaSource))  # one has to convert to list
        return randSource

    def inferFTSActivity(self, ftsOperation, rmsRequest, rmsOperation):
        """
        Tries to infer the FTS Activity
        """

        ### Data Challenge activity
        # All the tests with data challenges are done
        # on SE with '-DC-' in their name
        targetSEs = rmsOperation.targetSEList
        if any("-DC-" in se for se in targetSEs):
            return "Data Challenge"

        return super().inferFTSActivity(ftsOperation, rmsRequest, rmsOperation)
